"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

The Panoptes Context is one of the most important abstractions throughout the system. It is a thread-safe interface to
configuration and utilities throughout the system.

A Panoptes Context holds the system wide configuration, a logger and a Redis connection pool

In addition, a Context can optionally hold the following: multiple Key/Value stores, a Message Producer (with it's
underlying Kafka Client) and a ZooKeeper client

The Context object, once created, would be passed between multiple objects and methods within a process
"""
import os
import inspect
import logging
import re
from logging import StreamHandler, Formatter

import kazoo.client
import kazoo.client
import redis
from kafka import KafkaClient
from kafka.common import ConnectionError
from kazoo.exceptions import LockTimeout

from . import const
from .validators import PanoptesValidators
from .configuration_manager import PanoptesConfig
from .exceptions import PanoptesBaseException
from .utilities.helpers import get_calling_module_name
from .utilities.key_value_store import PanoptesKeyValueStore
from .utilities.message_queue import PanoptesMessageQueueProducer


class PanoptesContextError(PanoptesBaseException):
    """
    A class that encapsulates all context creation errors
    """
    pass


class PanoptesContextValidators(object):
    @classmethod
    def valid_panoptes_context(cls, panoptes_context):
        """
        valid_panoptes_context(cls, panoptes_context)

        Checks if the passed object is an instance of PanoptesContext

        Args:
            panoptes_context (PanoptesContext): The object to check

        Returns:
            bool: True if the object is not null and is an instance of PanoptesContext
        """
        return panoptes_context and isinstance(panoptes_context, PanoptesContext)


class PanoptesContext(object):
    """
    A thread-safe object that parses system wide config and sets up clients to various stores like Redis, Zookeeper and
    Kafka.

    A PanoptesContext is an essential object for all Panoptes subsystems. Creating a context does the following:

        * Parses and loads the system wide Panoptes configuration file
        * Creates a Python logger hierarchy based on the logging configuration file name provided in the system wide \
        configuration
        * A Redis Connection Pool (created through a call to the PanoptesConfiguraton class)
        * (Optional) Creates one or more Key/Value stores
        * (Optional) Creates a message producer client
        * (Optional) Creates a ZooKeeper client

    Args:
        config_file (str): Absolute path to the system wide configuration file
        key_value_store_class_list (list): A list of the Key/Value classes. PanoptesContext will create KV stores
        based on each class. Default is an empty list
        create_message_producer (bool): Whether a message producer client should be created. Default is False
        async_message_producer (bool): Whether the message producer client should be asynchronous. Default is False. \
        This parameter only matters if the create_message_producer is set to True
        create_zookeeper_client (bool): Whether a ZooKeeper client should be created


    Notes:
        * If a ZooKeeper client is created, three additional threads are created by by the Kazoo library
    """
    __rootLogger = None

    def __init__(self, config_file=None, key_value_store_class_list=None,
                 create_message_producer=False, async_message_producer=False, create_zookeeper_client=False):
        assert config_file is None or PanoptesValidators.valid_nonempty_string(config_file), \
            'config_file must be a non-empty string'
        assert key_value_store_class_list is None or isinstance(key_value_store_class_list,
                                                                list), 'key_value_store_class_list must be a list'

        self.__redis_connections = dict()
        self.__kv_stores = dict()
        self.__message_producer = None

        """
        Setup a default root logger so that in case configuration parsing or logger hierarchy creation fails, we have
        a place to send the error messages for failures
        """
        if not self.__class__.__rootLogger:
            try:
                self.__class__.__rootLogger = logging.getLogger(const.DEFAULT_ROOT_LOGGER_NAME)
                self.__class__.__rootLogger.setLevel(logging.INFO)
                handler = StreamHandler()
                handler.setFormatter(Formatter(fmt=const.DEFAULT_LOG_FORMAT))
                self.__class__.__rootLogger.addHandler(handler)
            except Exception as e:
                raise PanoptesContextError('Could not create root logger: %s' % str(e))

        if not config_file:
            if const.CONFIG_FILE_ENVIRONMENT_VARIABLE in os.environ:
                config_file = os.environ[const.CONFIG_FILE_ENVIRONMENT_VARIABLE]
            else:
                config_file = const.DEFAULT_CONFIG_FILE_PATH

        try:
            self.__logger = self.__class__.__rootLogger
            self.__config = self._get_panoptes_config(config_file)
            self.__logger = self._get_panoptes_logger()
        except Exception as e:
            raise PanoptesContextError('Could not create PanoptesContext: %s' % str(e))

        self.__redis_pool = self.get_redis_connection(const.DEFAULT_REDIS_GROUP_NAME)

        """
        Instantiate the KeyValueStore classes provide in the list (if any) and store the reference to the objects
        created in an object dictionary called __kv_stores
        """
        if key_value_store_class_list is not None:
            for key_value_store_class in key_value_store_class_list:
                if not inspect.isclass(key_value_store_class):
                    raise PanoptesContextError('Current item in key_value_store_class_list is not a class')
                if not issubclass(key_value_store_class, PanoptesKeyValueStore):
                    raise PanoptesContextError(key_value_store_class.__name__ +
                                               " in key_value_store_class_list does not subclass PanoptesKeyValueStore")

            for key_value_store_class in key_value_store_class_list:
                self.__kv_stores[key_value_store_class.__name__] = self._get_kv_store(key_value_store_class)

        if create_message_producer:
            self._kafka_client = self._get_kafka_client()

        if create_message_producer:
            self.__message_producer = self._get_message_producer(async_message_producer)

        if create_zookeeper_client:
            self.__zookeeper_client = self._get_zookeeper_client()

    def __repr__(self):
        kv_repr = 'KV Stores: [' + ','.join([str(Obj) for Obj in self.kv_stores]) + ']'
        config_repr = 'Config: ' + repr(self.config_object)
        redis_pool_repr = 'Redis pool set: ' + str(hasattr(self, '__redis_pool'))
        message_producer_repr = 'Message producer set: ' + str(hasattr(self, '__message_producer'))
        kafka_client_repr = 'Kafka client set: ' + str(hasattr(self, '_kafka_client'))
        zk_client_repr = 'Zookeeper client set: ' + str(hasattr(self, '__zookeeper_client'))
        return '[PanoptesContext: %s, %s, %s, %s, %s, %s]' \
               % (kv_repr, config_repr, redis_pool_repr, message_producer_repr, kafka_client_repr, zk_client_repr)

    def __del__(self):
        """
        Attempt to do resource clean-up when the reference count for PanoptesContext goes to zero, namely:
            * Delete any KV stores. Then delete __kv_stores.
            * Disconnect the redis pool.
            * Stop the message producer if one was requested/created.
            * Flush and close the kafka client.
            * Stop and close the zookeeper client if one was requested/created.
        """
        try:
            del self.__kv_stores
        except AttributeError as e:
            self.logger.error('__kv_stores attribute no longer exists: %s' % str(e))
        except Exception as e:
            self.logger.error('Attempt to delete _kv_stores failed: %s' % str(e))

        if hasattr(self, '_' + self.__class__.__name__ + '__message_producer'):
            try:
                self.__message_producer.stop()
            except Exception as e:
                self.logger.error('Attempt to stop message producer failed: %s' % str(e))

        if hasattr(self, '_kafka_client'):
            try:
                self._kafka_client.close()
            except Exception as e:
                self.logger.error('Attempt to close the Kafka client failed: %s' % str(e))

        if hasattr(self, '_' + self.__class__.__name__ + '__zookeeper_client'):
            try:
                self.__zookeeper_client.stop()
                self.__zookeeper_client.close()
            except Exception as e:
                self.logger.error('Attempt to stop and close the zookeeper client failed: %s' % str(e))

    def _get_panoptes_config(self, config_file):
        """
        Returns the system wide configuration to be used with the context

        Args:
            config_file (str): The path and name of the configuration file to parse

        Returns:
            PanoptesConfig: The Panoptes Config object that holds the system wide configuration

        Raises:
            PanoptesContextError: This exception is raised if any errors happen in reading or parsing the configuration
            file
        """
        self.__logger.info('Attempting to get Panoptes Configuration')
        try:
            panoptes_config = PanoptesConfig(self.__class__.__rootLogger, config_file)
        except Exception as e:
            raise PanoptesContextError('Could not get Panoptes Configuration object: %s' % str(e))

        self.__logger.info('Got Panoptes Configuration: %s' % panoptes_config)
        return panoptes_config

    def _get_panoptes_logger(self):
        """
        Returns the logger to be used by the context

        The method attempts to guess the name of the calling module based on introspection of the stack

        Returns:
            logger(logger): A Python logger subsystem logger

        Raises:
            PanoptesContextError: This exception is raised is any errors happen trying to instantiate the logger
        """
        self.__logger.info('Attempting to get logger')
        try:
            module = get_calling_module_name()
            logger = self.__rootLogger.getChild(module)
            self.__logger.info('Got logger for module %s' % module)
            return logger
        except Exception as e:
            raise PanoptesContextError('Could not get logger: %s' % str(e))

    def _get_redis_connection(self, group, shard):
        """
        Create and return a Redis Connection for the given group

        Returns:
            redis.StrictRedis: The Redis Connection

        Raises:
            Exception: Passes through any exceptions that happen in trying to get the connection pool
        """
        redis_group = self.__config.redis_urls_by_group[group][shard]

        self.__logger.info('Attempting to connect to Redis for group "{}", shard "{}", url "{}"'.format(group, shard,
                                                                                                        redis_group))

        redis_pool = redis.BlockingConnectionPool(host=redis_group.host,
                                                  port=redis_group.port,
                                                  db=redis_group.db,
                                                  password=redis_group.password)

        redis_connection = redis.StrictRedis(connection_pool=redis_pool)

        self.__logger.info('Successfully connected to Redis for group "{}", shard "{}", url "{}"'.format(group, shard,
                                                                                                         redis_group))

        return redis_connection

    def _get_kv_store(self, cls):
        """
        Create and return a Key/Value store

        Args:
            cls (class): The class of the Panoptes Key/Value store to create

        Returns:
            PanoptesKeyValueStore: The Key/Value store object created

        Raises:
            PanoptesContextError: Passes through any exceptions that happen in trying to create the Key/Value store
        """
        self.__logger.info('Attempting to connect to KV Store "{}"'.format(cls.__name__))
        try:

            key_value_store = cls(self)
        except Exception as e:
            raise PanoptesContextError('Could not connect to KV store "{}": {}'.format(cls.__name__, repr(e)))
        self.__logger.info('Connected to KV Store "{}": {}'.format(cls.__name__, key_value_store))
        return key_value_store

    def _get_kafka_client(self):
        """
        Create and return a Kafka Client

        Returns:
            KafkaClient: The created Kafka client

        Raises:
            PanoptesContextError: Passes through any exceptions that happen in trying to create the Kafka client
        """
        # The logic of the weird check that follows is this: KafkaClient initialization can fail if there is a problem
        # connecting with even one broker. What we want to do is: succeed if the client was able to connect to even one
        # broker. So, we catch the exception and pass it through - and then check the number of brokers connected to the
        # client in the next statement (if not kafka_client.brokers) and fail if the client is not connected to any
        # broker
        self.__logger.info('Attempting to connect Kafka')
        config = self.__config
        kafka_client = None
        try:
            kafka_client = KafkaClient(config.kafka_brokers)
        except ConnectionError:
            pass

        if not kafka_client.brokers:
            raise PanoptesContextError('Could not connect to any Kafka broker from this list: %s'
                                       % config.kafka_brokers)
        self.__logger.info('Successfully connected to Kafka brokers: %s' % kafka_client.brokers)

        return kafka_client

    def _get_message_producer(self, async):
        """
        Creates and returns a Message Producer

        Args:
            async (bool): Whether the created message producer should be asynchronous or not

        Returns:
            PanoptesMessageQueueProducer: The created message producer

        Raises:
            PanoptesContextError: asses through any exceptions that happen in trying to create the message producer
        """
        self.__logger.info('Attempting to connect to message bus')
        try:
            message_producer = PanoptesMessageQueueProducer(self, async)
        except Exception as e:
            raise PanoptesContextError('Could not connect to message bus: %s' % str(e))
        self.__logger.info('Connected to message bus: %s' % message_producer)
        return message_producer

    def _get_zookeeper_client(self):
        """
        Create and return a ZooKeeper client

        Returns:
            KazooClient: The created ZooKeeper client

        Raises:
            PanoptesContextError: Passes through any exceptions that happen in trying to create the ZooKeeper client
        """
        config = self.__config
        if not config.zookeeper_servers:
            raise PanoptesContextError('No Zookeeper servers configured')

        self.__logger.info('Attempting to connect to Zookeeper with servers: %s' % ",".join(config.zookeeper_servers))
        try:
            zk = kazoo.client.KazooClient(hosts=",".join(config.zookeeper_servers))
            zk.start()
        except Exception as e:
            raise PanoptesContextError('Could not connect to Zookeeper: %s' % str(e))
        self.__logger.info('Successfully connected to Zookeeper: %s' % zk)

        return zk

    def get_kv_store(self, key_value_store_class_name):
        """
        Get the Key Value Store object associated with the provided KeyValueStore class

        Args:
            key_value_store_class_name (class): The class (not just the classname string) for which the object is \
            desired

        Returns:
            KeyValueStore: An object which can be used to set/get values from the associated Key/Value store. Raises \
            PanoptesContextError if an object of the specified class does not exist
        """
        try:
            return self.__kv_stores[key_value_store_class_name.__name__]
        except KeyError:
            raise PanoptesContextError(
                'No Key Value Store based on class %s' % key_value_store_class_name)

    def get_redis_shard_count(self, group, fallback_to_default=True):
        try:
            return len(self.__config.redis_urls_by_group[group])
        except:
            if (group != const.DEFAULT_REDIS_GROUP_NAME) and fallback_to_default:
                return len(self.__config.redis_urls_by_group[const.DEFAULT_REDIS_GROUP_NAME])
            else:
                raise

    def get_redis_connection(self, group, shard=0, fallback_to_default=True):
        """
        Returns a Redis connection for the given group and shard

        Args:
            group (str): The name of the group for which to return the Redis connection
            shard (int): The number of the shard for which to return the Redis connection
            fallback_to_default (bool): If we can't find a connection for given group, whether to fallback to the \
            'default` group name

        Returns:
            redis.StrictRedis: The Redis connection
        """
        def _inner_get_redis_connection():
            try:
                connection = self._get_redis_connection(group, shard)
                self.__redis_connections[group][shard] = connection
            except:
                if (group != const.DEFAULT_REDIS_GROUP_NAME) and fallback_to_default:
                    self.__redis_connections[group][shard] = self.redis_pool
                else:
                    raise

        if group not in self.__redis_connections:
            self.__redis_connections[group] = dict()
            _inner_get_redis_connection()
        elif shard not in self.__redis_connections[group]:
            _inner_get_redis_connection()

        return self.__redis_connections[group][shard]

    def get_lock(self, path, timeout, retries=1, identifier=None, listener=None):
        """
        A wrapper around the kazoo library lock

        Args:
            path (str): A '/' separated path for the lock
            timeout (int): in seconds. Must be a positive integer
            retries (int): how many times to try before giving up. Zero implies try forever
            identifier (str): Name to use for this lock contender. This can be useful for querying \
            to see who the current lock contenders are
            listener (callable): The callable to use to handle Zookeeper state changes

        Returns:
            kazoo.recipe.lock.Lock: lock
        """
        assert isinstance(path, str) and re.search("^/\S+", path), 'path must be a non-empty string that begins with /'
        assert isinstance(timeout, int) and (timeout > 0), 'timeout must be a positive integer'
        assert isinstance(retries, int) and (retries > -1), 'retries must be a non-negative integer'
        assert identifier and isinstance(identifier, str), 'identifier must be a non-empty string'
        assert (not listener) or callable(listener), 'listener must be a callable'

        logger = self.logger
        calling_module = get_calling_module_name(2)
        logger.info("Creating lock for module: " + calling_module + " with lock parameters: path=" + path +
                    ",timeout=" + str(timeout) + ",retries=" + str(retries) + ",identifier=" + identifier)
        try:
            lock = self.zookeeper_client.Lock(path, identifier)
        except Exception as e:
            logger.error('Failed to create lock object: %s' % str(e))
            return None

        if retries == 0:
            while True:
                logger.info('Trying to acquire lock with client id "%s" under path %s. Other contenders: %s. '
                            % (identifier, path, lock.contenders()))
                try:
                    lock.acquire(timeout=timeout)
                except LockTimeout:
                    logger.info('Timed out after %d seconds trying to acquire lock.  Retrying.' % timeout)
                except Exception as e:
                    logger.info('Error in acquiring lock: %s.  Retrying.' % str(e))
                if lock.is_acquired:
                    break
        else:
            tries = 0
            while tries < retries:
                logger.info('Trying to acquire lock with client id "%s" under path %s. Other contenders: %s. ' %
                            (identifier, path, lock.contenders()))
                try:
                    lock.acquire(timeout=timeout)
                except LockTimeout:
                    logger.info('Timed out after %d seconds trying to acquire lock. Retrying %d more times' %
                                (timeout, retries - tries - 1))
                except Exception as e:
                    logger.info('Error in acquiring lock: %s.  Retrying %d more times' % (str(e), (retries - tries)))
                if lock.is_acquired:
                    break
                tries += 1
            if not lock.is_acquired:
                logger.warn('Unable to acquire lock after %d tries' % tries)

        if lock.is_acquired:
            logger.info(
                'Lock acquired. Other contenders: %s' % lock.contenders())

            if listener:
                self.zookeeper_client.add_listener(listener)

            return lock

    @property
    def config_object(self):
        """
        The PanoptesConfig object created by the context

        Returns:
            PanoptesConfig

        """
        return self.__config

    @property
    def config_dict(self):
        """
        A **copy** of the system wide configuration

        Returns:
            ConfigObj

        """
        return self.__config.get_config()

    @property
    def logger(self):
        """
        A module-aware logger which will try and guess the right name for the calling module

        Returns:
            logging.logger

        """
        return self.__logger

    @property
    def message_producer(self):
        """
        The message producer object which can be used to send messages

        Returns:
            PanoptesMessageQueueProducer

        """
        return self.__message_producer

    @property
    def redis_pool(self):
        """
        A Redis Connection Pool

        Returns:
            RedisConnectionPool

        """
        return self.__redis_pool

    @property
    def zookeeper_client(self):
        """
        A Kazoo ZooKeeper client

        Returns:
            KazooClient

        """
        return self.__zookeeper_client

    @property
    def kafka_client(self):
        """
        A Kafka client

        Returns:
            KafkaClient

        """
        return self._kafka_client

    @property
    def kv_stores(self):
        """
        Dictionary of KV stores

        Returns:
            A dictionary of KV store name/KV store class

        """
        return self.__kv_stores



import logging
import os
from unittest import TestCase

from mock import create_autospec

from yahoo_panoptes.framework.context import PanoptesContext
from yahoo_panoptes.framework.plugins.context import PanoptesPluginWithEnrichmentContext
from yahoo_panoptes.framework.resources import PanoptesResource
from yahoo_panoptes.framework.utilities.secrets import PanoptesSecretsStore
from yahoo_panoptes.framework.utilities.snmp.connection import PanoptesSNMPPluginConfiguration

panoptes_resource = PanoptesResource(
        resource_site='test_site',
        resource_class='test_class',
        resource_subclass='test_subclass',
        resource_type='test_type',
        resource_id='test_id',
        resource_endpoint='test_endpoint',
        resource_plugin='test_plugin'
)

plugin_conf = {
    'Core': {
        'name': 'Test Plugin',
        'module': 'test_plugin'
    },
    'main': {
        'execute_frequency': '60',
        'resource_filter': 'resource_class = "network"'
    }
}

path = os.path.dirname(os.path.realpath(__file__))
panoptes_context = PanoptesContext(config_file=os.path.join(path, 'config_files/test_panoptes_config.ini'))

secret_store = create_autospec(PanoptesSecretsStore, instance=True, spec_set=True)


class TestPanoptesSNMPPluginConfiguration(TestCase):
    def test_snmp_defaults(self):
        """Test that SNMP defaults from the Panoptes configuration file are used if no plugin specific SNMP config is
        provided and no site specific community secret is avalilable in the secrets store"""
        global secret_store

        secret_store.get_by_site.return_value = None

        plugin_context = create_autospec(
                PanoptesPluginWithEnrichmentContext, instance=True, spec_set=True,
                data=panoptes_resource,
                config=plugin_conf,
                snmp=panoptes_context.config_object.snmp_defaults,
                secrets=secret_store,
                logger=logging.getLogger(__name__)
        )

        snmp_configuration = PanoptesSNMPPluginConfiguration(plugin_context)

        self.assertEqual(snmp_configuration.port, 10161)
        self.assertEqual(snmp_configuration.proxy_port, 10161)
        self.assertEqual(snmp_configuration.connection_factory_module,
                         'yahoo_panoptes.framework.utilities.snmp.connection')
        self.assertEqual(snmp_configuration.connection_factory_class, 'PanoptesSNMPConnectionFactory')
        self.assertEqual(snmp_configuration.timeout, 5)
        self.assertEqual(snmp_configuration.retries, 1)
        self.assertEqual(snmp_configuration.non_repeaters, 0)
        self.assertEqual(snmp_configuration.max_repetitions, 25)
        self.assertEqual(snmp_configuration.community, 'public')
        self.assertEqual(snmp_configuration.community_string_key, 'snmp_community_string')

    def test_snmp_site_community(self):
        """
        Test that the site specific community string is used if present
        """
        global secret_store

        secret_store.get_by_site.return_value = 'test_site_community'

        plugin_context = create_autospec(
                PanoptesPluginWithEnrichmentContext, instance=True, spec_set=True,
                data=panoptes_resource,
                config=plugin_conf,
                snmp=panoptes_context.config_object.snmp_defaults,
                secrets=secret_store,
                logger=logging.getLogger(__name__)
        )

        snmp_configuration = PanoptesSNMPPluginConfiguration(plugin_context)
        self.assertEqual(snmp_configuration.community, 'test_site_community')

    def test_snmp_plugin_community(self):
        """
        Test that the plugin specific community string is used if present
        """
        global secret_store

        secret_store.get_by_site.return_value = 'test_site_community'

        plugin_conf_with_community = {
            'Core': {
                'name': 'Test Plugin',
                'module': 'test_plugin'
            },
            'main': {
                'execute_frequency': '60',
                'resource_filter': 'resource_class = "network"'
            },
            'snmp': {
                'community': 'test_plugin_community'
            }
        }

        plugin_context = create_autospec(
                PanoptesPluginWithEnrichmentContext, instance=True, spec_set=True,
                data=panoptes_resource,
                config=plugin_conf_with_community,
                snmp=panoptes_context.config_object.snmp_defaults,
                secrets=secret_store,
                logger=logging.getLogger(__name__)
        )

        snmp_configuration = PanoptesSNMPPluginConfiguration(plugin_context)
        self.assertEqual(snmp_configuration.community, 'test_plugin_community')

    def test_snmp_plugin_overrides(self):
        """
        Test that the plugin specific SNMP configuration is used if present
        """
        plugin_conf_with_overrides = {
            'Core': {
                'name': 'Test Plugin',
                'module': 'test_plugin'
            },
            'main': {
                'execute_frequency': '60',
                'resource_filter': 'resource_class = "network"'
            },
            'snmp': {
                'connection_factory_module': 'test_module',
                'connection_factory_class': 'test_class',
                'port': 162,
                'proxy_port': 10162,
                'timeout': 10,
                'retries': 2,
                'non_repeaters': 1,
                'max_repetitions': 10,
                'community_string_key': 'test_community_string_key'
            }
        }

        plugin_context = create_autospec(
                PanoptesPluginWithEnrichmentContext, instance=True, spec_set=True,
                data=panoptes_resource,
                config=plugin_conf_with_overrides,
                snmp=panoptes_context.config_object.snmp_defaults,
                secrets=secret_store,
                logger=logging.getLogger(__name__)
        )

        snmp_configuration = PanoptesSNMPPluginConfiguration(plugin_context)
        self.assertEqual(snmp_configuration.port, 162)
        self.assertEqual(snmp_configuration.proxy_port, 10162)
        self.assertEqual(snmp_configuration.connection_factory_module, 'test_module')
        self.assertEqual(snmp_configuration.connection_factory_class, 'test_class')
        self.assertEqual(snmp_configuration.timeout, 10)
        self.assertEqual(snmp_configuration.retries, 2)
        self.assertEqual(snmp_configuration.non_repeaters, 1)
        self.assertEqual(snmp_configuration.max_repetitions, 10)
        self.assertEqual(snmp_configuration.community_string_key, 'test_community_string_key')

    @staticmethod
    def _plugin_context_with_bad_snmp_configuration(**kwargs):
        plugin_conf_with_bad_configuration_value = {
            'Core': {
                'name': 'Test Plugin',
                'module': 'test_plugin'
            },
            'main': {
                'execute_frequency': '60',
                'resource_filter': 'resource_class = "network"'
            },
            'snmp': kwargs
        }

        return create_autospec(
                PanoptesPluginWithEnrichmentContext, instance=True, spec_set=True,
                data=panoptes_resource,
                config=plugin_conf_with_bad_configuration_value,
                snmp=panoptes_context.config_object.snmp_defaults,
                secrets=secret_store,
                logger=logging.getLogger(__name__)
        )

    def test_snmp_bad_configuration_values(self):
        """
        Test that 'bad' configuration values raise expected exceptions
        """
        with self.assertRaises(AssertionError):
            PanoptesSNMPPluginConfiguration(
                self._plugin_context_with_bad_snmp_configuration(connection_factory_module=0))
        with self.assertRaises(AssertionError):
            PanoptesSNMPPluginConfiguration(
                self._plugin_context_with_bad_snmp_configuration(connection_factory_module=''))

        with self.assertRaises(AssertionError):
            PanoptesSNMPPluginConfiguration(
                self._plugin_context_with_bad_snmp_configuration(connection_factory_class=0))
        with self.assertRaises(AssertionError):
            PanoptesSNMPPluginConfiguration(
                self._plugin_context_with_bad_snmp_configuration(connection_factory_class=''))

        with self.assertRaises(AssertionError):
            PanoptesSNMPPluginConfiguration(self._plugin_context_with_bad_snmp_configuration(community=0))
        with self.assertRaises(AssertionError):
            PanoptesSNMPPluginConfiguration(self._plugin_context_with_bad_snmp_configuration(community=''))

        with self.assertRaises(AssertionError):
            PanoptesSNMPPluginConfiguration(self._plugin_context_with_bad_snmp_configuration(community_string_key=0))
        with self.assertRaises(AssertionError):
            PanoptesSNMPPluginConfiguration(self._plugin_context_with_bad_snmp_configuration(community_string_key=''))

        with self.assertRaises(ValueError):
            PanoptesSNMPPluginConfiguration(self._plugin_context_with_bad_snmp_configuration(port=''))
        with self.assertRaises(AssertionError):
            PanoptesSNMPPluginConfiguration(self._plugin_context_with_bad_snmp_configuration(port=-1))
        with self.assertRaises(AssertionError):
            PanoptesSNMPPluginConfiguration(self._plugin_context_with_bad_snmp_configuration(port=65536))

        with self.assertRaises(ValueError):
            PanoptesSNMPPluginConfiguration(self._plugin_context_with_bad_snmp_configuration(proxy_port=''))
        with self.assertRaises(AssertionError):
            PanoptesSNMPPluginConfiguration(self._plugin_context_with_bad_snmp_configuration(proxy_port=-1))
        with self.assertRaises(AssertionError):
            PanoptesSNMPPluginConfiguration(self._plugin_context_with_bad_snmp_configuration(proxy_port=65536))

        with self.assertRaises(ValueError):
            PanoptesSNMPPluginConfiguration(self._plugin_context_with_bad_snmp_configuration(timeout=''))
        with self.assertRaises(AssertionError):
            PanoptesSNMPPluginConfiguration(self._plugin_context_with_bad_snmp_configuration(timeout=0))

        with self.assertRaises(ValueError):
            PanoptesSNMPPluginConfiguration(self._plugin_context_with_bad_snmp_configuration(retries=''))
        with self.assertRaises(AssertionError):
            PanoptesSNMPPluginConfiguration(self._plugin_context_with_bad_snmp_configuration(retries=0))

        with self.assertRaises(ValueError):
            PanoptesSNMPPluginConfiguration(self._plugin_context_with_bad_snmp_configuration(non_repeaters=''))
        with self.assertRaises(AssertionError):
            PanoptesSNMPPluginConfiguration(self._plugin_context_with_bad_snmp_configuration(non_repeaters=-1))

        with self.assertRaises(ValueError):
            PanoptesSNMPPluginConfiguration(self._plugin_context_with_bad_snmp_configuration(max_repetitions=''))
        with self.assertRaises(AssertionError):
            PanoptesSNMPPluginConfiguration(self._plugin_context_with_bad_snmp_configuration(max_repetitions=0))



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.
"""

import json
import time
import unittest

from mock import *

from test_helpers import ordered
from yahoo_panoptes.framework.metrics import PanoptesMetricsGroup, PanoptesMetricDimension, \
    PanoptesMetricsGroupSet, PanoptesMetric, PanoptesMetricSet, PanoptesMetricType, PanoptesMetricsGroupEncoder, \
    METRICS_TIMESTAMP_PRECISION
from yahoo_panoptes.framework.resources import PanoptesResource

mock_time = MagicMock()
mock_time.return_value = round(1538082314.09, METRICS_TIMESTAMP_PRECISION)


class TestMetrics(unittest.TestCase):
    def setUp(self):
        self.__panoptes_resource = PanoptesResource(resource_site='test', resource_class='test',
                                                    resource_subclass='test',
                                                    resource_type='test', resource_id='test', resource_endpoint='test',
                                                    resource_plugin='test')
        self.__panoptes_resource.add_metadata('test', 'test')

    def test_panoptes_metric(self):
        with self.assertRaises(AssertionError):
            PanoptesMetric(None, 0, PanoptesMetricType.GAUGE)

        with self.assertRaises(ValueError):
            PanoptesMetric('1', 0, PanoptesMetricType.GAUGE)

        with self.assertRaises(AssertionError):
            PanoptesMetric('test_metric', None, PanoptesMetricType.GAUGE)

        with self.assertRaises(AssertionError):
            PanoptesMetric('test_metric', 0, None)

        with self.assertRaises(AssertionError):
            PanoptesMetric('test_metric', True, PanoptesMetricType.GAUGE)

        metric1 = PanoptesMetric('test_metric', 0, PanoptesMetricType.GAUGE,
                                 metric_creation_timestamp=mock_time.return_value)

        self.assertEqual(metric1.metric_name, 'test_metric')
        self.assertEqual(metric1.metric_value, 0)
        self.assertEqual(metric1.metric_timestamp, mock_time.return_value)
        self.assertEqual(metric1.metric_type, PanoptesMetricType.GAUGE)
        self.assertEqual(repr(metric1),
                         "{{'metric_creation_timestamp': {}, 'metric_type': 'gauge', 'metric_name': 'test_metric', "
                         "'metric_value': 0}}".format(mock_time.return_value))

        self.assertNotEqual(metric1, None)

        # Check PanoptesMetric.__eq__
        assert metric1 == PanoptesMetric('test_metric', 0, PanoptesMetricType.GAUGE)
        with self.assertRaises(AssertionError):
            assert metric1 == PanoptesMetricDimension("test", "value")
        with self.assertRaises(AssertionError):
            assert metric1 == PanoptesMetric('different_name', 0, PanoptesMetricType.GAUGE)
        with self.assertRaises(AssertionError):
            assert metric1 == PanoptesMetric('test_metric', 1, PanoptesMetricType.GAUGE)
        with self.assertRaises(AssertionError):
            assert metric1 == PanoptesMetric('test_metric', 0, PanoptesMetricType.COUNTER)

    @patch('yahoo_panoptes.framework.metrics.time', mock_time)
    def test_panoptes_metric_json_and_repr(self):
        metric = PanoptesMetric('test_metric', 0, PanoptesMetricType.GAUGE, mock_time.return_value)
        serialized = json.loads(metric.json)
        expected = {"metric_creation_timestamp": mock_time.return_value,
                    "metric_name": "test_metric",
                    "metric_type": "gauge",
                    "metric_value": 0}

        self.assertEqual(ordered(serialized), ordered(expected))

    def testMetricsGroup(self):
        now = round(time.time(), METRICS_TIMESTAMP_PRECISION)
        metrics_group = PanoptesMetricsGroup(self.__panoptes_resource, 'test', 120)
        self.assertEqual(metrics_group.group_type, 'test')
        self.assertEqual(metrics_group.interval, 120)
        self.assertEqual(metrics_group.schema_version, '0.2')
        self.assertGreaterEqual(metrics_group.creation_timestamp, now)

        with patch('yahoo_panoptes.framework.metrics.time', mock_time):
            metrics_group = PanoptesMetricsGroup(self.__panoptes_resource, 'test', 120)

            dimension_one = PanoptesMetricDimension('if_alias', 'bar')
            dimension_two = PanoptesMetricDimension('if_alias', 'foo')

            metrics_group.add_dimension(dimension_one)

            with self.assertRaises(KeyError):
                metrics_group.add_dimension(dimension_two)

            #  Test basic dimension operations
            self.assertEqual(len(metrics_group.dimensions), 1)
            self.assertTrue(metrics_group.contains_dimension_by_name('if_alias'))
            self.assertFalse(metrics_group.contains_dimension_by_name('baz'))
            self.assertEqual(metrics_group.get_dimension_by_name('if_alias').value, 'bar')
            metrics_group.delete_dimension_by_name('if_alias')
            self.assertFalse(metrics_group.contains_dimension_by_name('if_alias'))
            self.assertEqual(len(metrics_group.dimensions), 0)
            self.assertEqual(metrics_group.get_dimension_by_name('foo'), None)

            metrics_group.add_dimension(dimension_two)
            dimension_three = PanoptesMetricDimension('if_alias', 'test')
            metrics_group.upsert_dimension(dimension_three)
            self.assertEqual(len(metrics_group.dimensions), 1)
            self.assertEqual(metrics_group.get_dimension_by_name('if_alias').value, 'test')
            dimension_four = PanoptesMetricDimension('if_name', 'eth0')
            metrics_group.upsert_dimension(dimension_four)
            self.assertEqual(len(metrics_group.dimensions), 2)

            #  Test basic metric operations
            with self.assertRaises(AssertionError):
                metrics_group.add_metric(None)

            metric = PanoptesMetric('test_metric', 0, PanoptesMetricType.GAUGE)
            metrics_group.add_metric(metric)
            with self.assertRaises(KeyError):
                metrics_group.add_metric(metric)

            to_json = metrics_group.json
            metrics = PanoptesMetricsGroup.flatten_metrics(json.loads(to_json)['metrics'])
            self.assertEquals(metrics['gauge']['test_metric']['value'], 0)

            metrics_group_two = PanoptesMetricsGroup(self.__panoptes_resource, 'test', 120)
            metrics_group_two.add_dimension(dimension_two)
            metrics_group_two.upsert_dimension(dimension_three)
            metrics_group_two.upsert_dimension(dimension_four)
            metrics_group_two.add_metric(metric)

            self.assertEqual(metrics_group, metrics_group_two)

            # Check PanoptesMetricsGroup.__eq__
            panoptes_resource_two = PanoptesResource(resource_site='test2', resource_class='test2',
                                                     resource_subclass='test2',
                                                     resource_type='test2', resource_id='test2',
                                                     resource_endpoint='test2',
                                                     resource_plugin='test2')

            metrics_group_two = PanoptesMetricsGroup(panoptes_resource_two, 'test', 120)
            metrics_group_three = PanoptesMetricsGroup(self.__panoptes_resource, 'test', 120)
            with self.assertRaises(AssertionError):
                assert metrics_group_two == metrics_group_three

            metrics_group_three = metrics_group.copy()

            with self.assertRaises(AssertionError):
                assert metrics_group == dimension_one
            assert metrics_group == metrics_group_three

            metrics_group_three.delete_dimension_by_name("if_name")
            with self.assertRaises(AssertionError):
                assert metrics_group == metrics_group_three
            metrics_group_three.upsert_dimension(dimension_four)
            assert metrics_group == metrics_group_three

            metric_two = PanoptesMetric('test_metric_2', 1, PanoptesMetricType.GAUGE)
            metrics_group_three.add_metric(metric_two)
            with self.assertRaises(AssertionError):
                assert metrics_group == metrics_group_three

            #  Test PanoptesMetricsGroup.__repr__
            _METRICS_GROUP_REPR = "{{'metrics_group_interval': 120, " \
                                  "'resource': plugin|test|site|test|class|test|subclass|test|type|test|id|test|" \
                                  "endpoint|test, 'dimensions': set([{{'dimension_name': 'if_alias', " \
                                  "'dimension_value': 'test'}}, " \
                                  "{{'dimension_name': 'if_name', 'dimension_value': 'eth0'}}]), " \
                                  "'metrics_group_type': 'test', " \
                                  "'metrics': set([{{'metric_creation_timestamp': {}, " \
                                  "'metric_type': 'gauge', 'metric_name': 'test_metric', 'metric_value': 0}}]), " \
                                  "'metrics_group_creation_timestamp': {}, " \
                                  "'metrics_group_schema_version': '0.2'}}".format(mock_time.return_value,
                                                                                   mock_time.return_value)
            self.assertEqual(repr(metrics_group), _METRICS_GROUP_REPR)

            dimensions_as_dicts = [{'dimension_name': dimension.name,
                                    'dimension_value': dimension.value} for dimension in metrics_group.dimensions]
            self.assertEqual(PanoptesMetricsGroup.flatten_dimensions(dimensions_as_dicts),
                             {'if_alias': 'test', 'if_name': 'eth0'})

    def test_panoptes_metrics_group_encoder(self):
        test_dict = dict()
        encoder = PanoptesMetricsGroupEncoder()

        mock_default = Mock(json.JSONEncoder.default)
        with patch('yahoo_panoptes.framework.metrics.json.JSONEncoder.default', mock_default):
            encoder.default(test_dict)
            mock_default.assert_called_once()

    def test_panoptes_metric_dimension(self):
        with self.assertRaises(ValueError):
            PanoptesMetricDimension('contain$_invalid_character$', 'bar')
        with self.assertRaises(ValueError):
            PanoptesMetricDimension('foo', 'contains_pipe|')

        dimension_one = PanoptesMetricDimension('if_alias', 'bar')

        self.assertEqual(dimension_one.json, '{"dimension_name": "if_alias", "dimension_value": "bar"}')
        self.assertEqual(repr(dimension_one), "{'dimension_name': 'if_alias', 'dimension_value': 'bar'}")

        metric_one = PanoptesMetric('test_metric', 0, PanoptesMetricType.GAUGE,
                                    metric_creation_timestamp=mock_time.return_value)

        with self.assertRaises(AssertionError):
            assert dimension_one == metric_one

        dimension_two = PanoptesMetricDimension('if_alias', 'foo')
        with self.assertRaises(AssertionError):
            assert dimension_one == dimension_two

        dimension_three = PanoptesMetricDimension('if_alias', 'bar')
        assert dimension_one == dimension_three

    def test_metrics_group_hash(self):
        now = round(time.time(), METRICS_TIMESTAMP_PRECISION)
        metrics_group = PanoptesMetricsGroup(self.__panoptes_resource, 'test', 120)
        metrics_group_two = PanoptesMetricsGroup(self.__panoptes_resource, 'test', 120)

        dimension = PanoptesMetricDimension('if_alias', 'bar')
        metric = PanoptesMetric('test_metric', 0, PanoptesMetricType.GAUGE, metric_creation_timestamp=now)
        metric_diff_timestamp = PanoptesMetric('test_metric', 0, PanoptesMetricType.GAUGE,
                                               metric_creation_timestamp=now + 0.01)

        metrics_group.add_dimension(dimension)
        metrics_group_two.add_dimension(dimension)

        self.assertEqual(metrics_group.__hash__(), metrics_group_two.__hash__())

        metrics_group.add_metric(metric)
        metrics_group_two.add_metric(metric_diff_timestamp)

        self.assertEqual(metrics_group.__hash__(), metrics_group_two.__hash__())

    @patch('yahoo_panoptes.framework.metrics.time', mock_time)
    def test_panoptes_metrics_group_set(self):
        """Tests basic PanoptesMetricsGroupSet operations"""
        metrics_group_set = PanoptesMetricsGroupSet()
        metrics_group = PanoptesMetricsGroup(self.__panoptes_resource, 'test', 120)
        metrics_group_two = PanoptesMetricsGroup(self.__panoptes_resource, 'test', 120)
        metrics_group_set.add(metrics_group)
        metrics_group_set.add(metrics_group_two)
        assert len(metrics_group_set) == 1
        self.assertIn(metrics_group, metrics_group_set.metrics_groups)

        metrics_group_set.remove(metrics_group_two)
        assert len(metrics_group_set) == 0

        metrics_group_set.add(metrics_group)
        metrics_group_three = PanoptesMetricsGroup(self.__panoptes_resource, 'test3', 120)
        metrics_group_three.add_metric(PanoptesMetric("test3", 0.0, PanoptesMetricType.GAUGE))
        metrics_group_set.add(metrics_group_three)
        assert len(metrics_group_set) == 2

        metrics_group_set_two = PanoptesMetricsGroupSet()
        metrics_group_four = PanoptesMetricsGroup(self.__panoptes_resource, 'test', 120)
        metrics_group_four.add_metric(PanoptesMetric("test4", 0.0, PanoptesMetricType.GAUGE))
        metrics_group_set_two.add(metrics_group_four)
        assert len(metrics_group_set_two) == 1

        #  Test PanoptesMetricsGroupSet.__add__
        metrics_group_set_union = metrics_group_set + metrics_group_set_two
        assert len(metrics_group_set_union) == 3

        with self.assertRaises(AssertionError):
            metrics_group_set.remove(self.__panoptes_resource)

        with self.assertRaises(TypeError):
            metrics_group_set + metrics_group

        #  Test PanoptesMetricsGroupSet.__iter__ & 'next'
        metrics_group_count = 0
        metrics_group_set_union_interator = iter(metrics_group_set_union)
        for _ in metrics_group_set_union:
            self.assertIn(metrics_group_set_union_interator.next(), metrics_group_set_union.metrics_groups)
            metrics_group_count += 1
        assert len(metrics_group_set_union) == metrics_group_count
        with self.assertRaises(Exception):
            metrics_group_set_union_interator.next()

        #  Test PanoptesMetricsGroupSet.__repr__
        _METRICS_GROUP_SET_REPR = "set([{{'metrics_group_interval': 120, " \
                                  "'resource': plugin|test|site|test|class|test|subclass|test|type|test|id|test|" \
                                  "endpoint|test, 'dimensions': set([]), 'metrics_group_type': 'test', " \
                                  "'metrics': set([]), 'metrics_group_creation_timestamp': {}, " \
                                  "'metrics_group_schema_version': '0.2'}}, {{'metrics_group_interval': 120, " \
                                  "'resource': plugin|test|site|test|class|test|subclass|test|type|test|id|test|" \
                                  "endpoint|test, 'dimensions': set([]), 'metrics_group_type': 'test3', " \
                                  "'metrics': set([{{'metric_creation_timestamp': {}, " \
                                  "'metric_type': 'gauge', " \
                                  "'metric_name': 'test3', 'metric_value': 0.0}}]), " \
                                  "'metrics_group_creation_timestamp': {}, " \
                                  "'metrics_group_schema_version': '0.2'}}])".format(mock_time.return_value,
                                                                                     mock_time.return_value,
                                                                                     mock_time.return_value)
        self.assertEqual(repr(metrics_group_set), _METRICS_GROUP_SET_REPR)

    def test_panoptes_metric_set(self):
        metric_set = PanoptesMetricSet()
        metric1 = PanoptesMetric('test_metric', 0, PanoptesMetricType.GAUGE,
                                 metric_creation_timestamp=mock_time.return_value)
        metric2 = PanoptesMetric('test_metric', 0, PanoptesMetricType.GAUGE,
                                 metric_creation_timestamp=mock_time.return_value)

        metric_set.add(metric1)
        metric_set.add(metric2)
        assert len(metric_set) == 1

        self.assertIn(metric1, metric_set.metrics)

        #  Test PanoptesMetricSet.__repr__
        _METRIC_SET_REPR = "set([{{'metric_creation_timestamp': {}, 'metric_type': 'gauge', " \
                           "'metric_name': 'test_metric', 'metric_value': 0}}])".format(mock_time.return_value)
        self.assertEqual(repr(metric_set), _METRIC_SET_REPR)

        with self.assertRaises(Exception):
            metric_set.remove(self.__panoptes_resource)

        metric_set.remove(metric1)
        assert len(metric_set) == 0

        #  Test PanoptesMetricSet.__iter__ and 'next'
        metric_count = 0
        metric_set_iterator = iter(metric_set)
        for _ in metric_set:
            self.assertIn(metric_set_iterator.next(), metric_set.metrics)
            metric_count += 1
        assert len(metric_set) == metric_count
        with self.assertRaises(Exception):
            metric_set_iterator.next()

if __name__ == '__main__':
    unittest.main()



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This module implements Interface EnrichmentGroup defined with schema validator
"""

from ...framework.enrichment import PanoptesEnrichmentSchemaValidator, PanoptesEnrichmentGroup

INTERFACE_SCHEMA_NAMESPACE = 'interface'


class PanoptesInterfaceEnrichmentSchemaValidator(PanoptesEnrichmentSchemaValidator):
    schema = {
        'enrichment_label': {
            'type': 'dict',
            'schema': {
                'description': {'type': 'string', 'required': True},
                'media_type': {'type': 'string', 'required': True},
                'interface_name': {'type': 'string', 'required': True},
                'alias': {'type': 'string', 'required': True},
                'configured_speed': {'type': 'integer', 'required': True},
                'port_speed': {'type': 'integer', 'required': True},
                'parent_interface_name': {'type': 'string', 'required': True},
                'parent_interface_media_type': {'type': 'string', 'required': True},
                'parent_interface_configured_speed': {'type': 'integer', 'required': True},
                'parent_interface_port_speed': {'type': 'integer', 'required': True},
                'remote_device': {'type': 'string', 'required': False},
                'remote_device_site': {'type': 'string', 'required': False},
                'remote_device_function': {'type': 'string', 'required': False},
                'remote_device_purpose': {'type': 'string', 'required': False},
                'physical_address': {'type': 'string', 'required': False}
            }
        }
    }


class PanoptesInterfaceEnrichmentGroup(PanoptesEnrichmentGroup):
    def __init__(self, enrichment_ttl, execute_frequency):
        super(PanoptesInterfaceEnrichmentGroup, self).__init__(
            namespace=INTERFACE_SCHEMA_NAMESPACE,
            schema_validator=PanoptesInterfaceEnrichmentSchemaValidator(),
            enrichment_ttl=enrichment_ttl,
            execute_frequency=execute_frequency)



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This module contains custom validators which can be used in assert checks
"""
import os
import time
import numbers
import re
import logging

from six import string_types, integer_types

NUMERIC_OID = re.compile(r'\.(\d+)(\.\d+)*')


class PanoptesValidators(object):
    """
    This class contains custom validator methods
    """

    @classmethod
    def valid_plugin_result_class(cls, plugin_result_class):
        """
        valid_plugin_result_class(cls, plugin_result_class)

        Checks if the passed class is a subclass of object class
        Args:
            plugin_result_class (object): The class to check

        Returns:
            bool: True if the class is not null and is an subclass of object
        """
        return plugin_result_class and issubclass(plugin_result_class, object)

    @classmethod
    def valid_logger(cls, logger):
        return isinstance(logger, logging.Logger)

    @classmethod
    def valid_callback(cls, callback):
        """
        valid_callback(cls, callback)

        Checks if the passed object is a callable
        Args:
            callback (callback): The object to check

        Returns:
            bool: True if the object passed is a callable

        """
        return callback and hasattr(callback, '__call__')

    @classmethod
    def valid_hashable_object(cls, object_to_check):
        """
        Checks if the passed object is hashable

        Args:
            object_to_check (object): The object to check

        Returns:
            bool: True if the object is hashable
        """

        return hasattr(object_to_check, '__hash__') and object_to_check.__hash__

    @classmethod
    def valid_port(cls, port):
        return type(port) in integer_types and (port > 0) and (port < 65536)

    @classmethod
    def valid_number(cls, value):
        return isinstance(value, numbers.Number) and not isinstance(value, bool)

    @classmethod
    def valid_nonzero_integer(cls, value):
        return type(value) in integer_types and (value > 0)

    @classmethod
    def valid_positive_integer(cls, value):
        """
        N.b. In Panoptes, '0' is considered a positive integer.
        """
        return type(value) in integer_types and (value > -1)

    @classmethod
    def valid_nonempty_string(cls, value):
        return isinstance(value, string_types) and (len(value) > 0)

    @classmethod
    def valid_readable_path(cls, path):
        try:
            return True if (os.path.isdir(path) and os.access(path, os.R_OK)) else False
        except:
            return False

    @classmethod
    def valid_readable_file(cls, filename):
        try:
            return True if (os.path.isfile(filename) and os.access(filename, os.R_OK)) else False
        except:
            return False

    @classmethod
    def valid_nonempty_iterable_of_strings(cls, value):
        return hasattr(value, '__iter__') and \
               hasattr(value, '__len__') and \
               len(value) > 0 and \
               all(type(val) == str for val in value) and \
               all(len(val) > 0 for val in value)

    @classmethod
    def valid_timestamp(cls, timestamp, max_age=604800, max_skew=60):
        """
        Checks is the passed value is a valid unix timestamp, which is not older than max_age (defaults to 7 days) and\
        not more than max_skew (defaults to 60 seconds)

        max_skew is needed to handle distributed systems not have exactly synced clocks

        Args:
            timestamp (int, float): An integer or float which contains the timestamp
            max_age (int, float): The maximum age, in seconds, of how old the timestamp can be from the current time
            max_skew (int, float): The maximum skew (seconds in the future) to how much more the timestamp can be from\
            the current time

        Returns:
            bool: True if the validation passes, false otherwise

        """
        return type(timestamp) in [int, float] and \
            type(max_age) in [int, float] and \
            type(max_skew) in [int, float] and \
            (time.time() - max_age) <= timestamp <= (time.time() + max_skew)

    @classmethod
    def valid_numeric_snmp_oid(cls, oid):
        return True if NUMERIC_OID.match(oid) else False



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This module provides convenience classes to interact with Celery: there are classes for representing Celery
Configuration, Instances and an in-memory Scheduler
"""
import heapq
import threading

from celery import Celery
from celery.beat import Scheduler, event_t

from . import const
from .context import PanoptesContext
from .exceptions import PanoptesBaseException
from .validators import PanoptesValidators

thread_lock = threading.Lock()


class PanoptesCeleryError(PanoptesBaseException):
    """
    The exception class for Panoptes Celery errors
    """
    pass


class PanoptesCeleryConfig(object):
    """
    The base Celery Config class for Panoptes

    This class would only contain the attributes used to configure a Celery app
    """
    celery_accept_content = ['application/json', 'json']
    worker_prefetch_multiplier = 1
    task_acks_late = True

    def __init__(self, app_name):
        assert PanoptesValidators.valid_nonempty_string(app_name), 'app_name must be a non-empty string'
        self._celery_app_name = app_name

    @property
    def app_name(self):
        """
        The Celery App name

        Returns:
            str: The Celery App name
        """
        return self._celery_app_name


class PanoptesCeleryValidators(object):
    @classmethod
    def valid_celery_config(cls, celery_config):
        """
        valid_celery_config(cls, celery_config)

        Checks if the passed object is an instance of PanoptesCeleryConfig

        Args:
            celery_config (PanoptesCeleryConfig): The object to check

        Returns:
            bool: True if the object is not null and is an instance of PanoptesCeleryConfig
        """
        return celery_config and isinstance(celery_config, PanoptesCeleryConfig)


class PanoptesCeleryInstance(object):
    """
    Create a Celery instance

    Args:
        panoptes_context (PanoptesContext): The PanoptesContext to use. The system wide configuration associated with \
        the context would be used to create the Celery application - specially, the celery_broker would be set to the \
        redis_client from the configuration attached to the context
        celery_config (PanoptesCeleryConfig): The class containing attributes for the Celery app

    Returns:
        None
    """

    def __init__(self, panoptes_context, celery_config):
        assert isinstance(panoptes_context, PanoptesContext), 'panoptes_context must be an instance of PanoptesContext'
        assert isinstance(celery_config,
                          PanoptesCeleryConfig), 'celery_config must be an instance of PanoptesCeleryConfig'
        logger = panoptes_context.logger
        celery_broker = panoptes_context.config_object.redis_urls_by_group[const.DEFAULT_REDIS_GROUP_NAME]
        logger.info('Creating Celery Application "%s" with broker "%s"' % (
            celery_config.app_name, celery_broker))
        try:
            # TODO: The '0' after celery_broker below refers to the first shard
            self.__celery_instance = Celery(celery_config.app_name, broker=celery_broker[0].url)
        except Exception as e:
            logger.error('Failed to create Celery Application: %s' % str(e))

        self.__celery_instance.config_from_object(celery_config)
        logger.info('Created Celery Application: %s' % self.__celery_instance)

    @property
    def celery(self):
        """
        Returns the instance of the Celery app

        Returns:
            Celery
        """
        return self.__celery_instance


class PanoptesCeleryPluginScheduler(Scheduler):
    """
    The base plugin scheduler class in Panoptes
    """
    def update(self, logger, new_schedule):
        """
        Updates the currently installed scheduled

        Args:
            logger (logging.logger): The logger to use
            new_schedule (dict): The new schedule
        Returns:
            None
        """
        logger.debug('New schedule: %s' % str(new_schedule))
        logger.info('Going to schedule %d tasks' % len(new_schedule))
        with thread_lock:
            self.merge_inplace(new_schedule)
            logger.info('Going to add default entries')
            self.install_default_entries(new_schedule)
        logger.info('Scheduler now has %d tasks' % len(self.schedule))

    def tick(self, event_t=event_t, min=min, heappop=heapq.heappop, heappush=heapq.heappush):
        """
        Make the tick function thread safe
        """
        with thread_lock:
            super(PanoptesCeleryPluginScheduler, self).tick(event_t=event_t,
                                                            min=min,
                                                            heappop=heapq.heappop,
                                                            heappush=heapq.heappush)



# Taken from https://github.com/wkeeling/ratelimitingfilter because it's not available in PyPi as a package
from collections import defaultdict
import difflib
from functools import partial
import logging
import os
from time import time


class RateLimitingFilter(logging.Filter):
    """
    The RateLimitingFilter is a logging filter that can be used to throttle the number
    of records being sent through a logging handler. Internally the RateLimitingFilter is
    based on an implementation of the the Token Bucket algorithm to restrict the
    throughput of records. The desired throughput can be configured when instantiating
    an instance of the filter.
    """

    def __init__(self, rate=1, per=30, burst=1, **kwargs):
        """
        Create an instance of the RateLimitingFilter allowing a default rate of 1 record
        every 30 seconds when no arguments are supplied.

        :param rate: The number of records to restrict to, per the specified time interval. Default 1.
        :param per: The number of seconds during which 'rate' records may be sent. Default 30.
        :param burst: The maximum number of records that can be sent before rate limiting kicks in.
        :param kwargs: Additional config options that can be passed to the filter.
        """
        super(RateLimitingFilter, self).__init__()

        self.rate = rate
        self.per = per
        self.burst = burst

        self._default_bucket = TokenBucket(rate, per, burst)
        self._substr_buckets = None
        self._auto_buckets = None

        if 'match' in kwargs:
            if kwargs['match'] == 'auto':
                self._auto_buckets = defaultdict(partial(TokenBucket, rate, per, burst))
            else:
                self._substr_buckets = {}
                for s in kwargs['match']:
                    # Create a new token bucket for each substring
                    self._substr_buckets[s] = TokenBucket(rate, per, burst)

    def filter(self, record):
        """
        Determines whether the supplied record should be logged based upon current rate limits.

        If rate limits have been reached, the record is not logged and a counter is incremented
        to indicate that the record has been throttled. The next time a record is successfully
        logged, the number of previously throttled records is appended to the end of the message.

        :param record: The record to log.
        :return: True if the record can be logged, False otherwise.
        """

        bucket = self._bucket_for(record)

        if not bucket:
            return True

        if bucket.consume():
            if bucket.limited > 0:
                # Append a message to the record indicating the number of previously suppressed messages
                record.msg += '{linesep}... {num} additional messages suppressed'.format(linesep=os.linesep,
                                                                                         num=bucket.limited)
            bucket.limited = 0
            return True

        # Rate limit
        bucket.limited += 1
        return False

    def _bucket_for(self, record):
        if self._substr_buckets is not None:
            return self._get_substr_bucket(record)
        elif self._auto_buckets is not None:
            return self._get_auto_bucket(record)

        return self._default_bucket

    def _get_substr_bucket(self, record):
        # Locate the relevant token bucket by matching the configured substrings against the message
        for substr in self._substr_buckets:
            if substr in record.msg:
                return self._substr_buckets[substr]

        return None  # None indicates no filtering

    def _get_auto_bucket(self, record):
        if record.msg in self._auto_buckets:
            # We have an exact match - there is a bucket configured for this message
            return self._auto_buckets[record.msg]

        # Check whether we have a partial match - whether part of the message
        # matches against a token bucket we previously created for a similar message
        for msg, bucket in self._auto_buckets.items():
            matcher = difflib.SequenceMatcher(None, msg, record.msg)
            if matcher.ratio() >= 0.75:  # Might want to make the ratio threshold configurable?
                return bucket

        # No match, so create a new bucket for this message
        return self._auto_buckets[record.msg]


class TokenBucket(object):
    """
    An implementation of the Token Bucket algorithm.
    """

    def __init__(self, rate, per, burst):
        self._rate = rate
        self._per = per or 1
        self._burst = burst
        self._allowance = burst
        self._last_check = time()
        self.limited = 0

    def consume(self):
        """
        Attempts to consume a token from the bucket based upon the current state of the bucket.

        :return: True if a token can be consumed, False otherwise.
        """
        now = time()
        delta = now - self._last_check
        self._last_check = now

        self._allowance += delta * (self._rate / self._per)

        if self._allowance > self._burst:
            self._allowance = self._burst

        if self._allowance < 1:
            return False

        self._allowance -= 1
        return True

    def __repr__(self):
        return 'TokenBucket(rate={0._rate}, per={0._per}, burst={0._burst})'.format(self)



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This module implements the Enrichment Plugin Agent which accepts plugin names as Celery Task parameters and executes
them. Results, if any, returned by each plugin are expected to be PanoptesEnrichmentGroupSet. If the returned Metrics
Set is not empty, each metric from the metric set is placed on a Kafka queue named 'metric'

This module is expected to be imported and executed though the Celery 'worker' command line tool
"""
import faulthandler
import sys

from celery import task

from ..framework import const
from ..framework.exceptions import PanoptesBaseException
from ..framework.context import PanoptesContext
from ..framework.celery_manager import PanoptesCeleryConfig, PanoptesCeleryInstance
from ..framework.resources import PanoptesResourceStore, PanoptesResourcesKeyValueStore
from ..framework.plugins.panoptes_base_plugin import PanoptesPluginInfo
from ..framework.plugins.runner import PanoptesPluginWithEnrichmentRunner
from ..framework.utilities.helpers import get_calling_module_name
from ..framework.utilities.key_value_store import PanoptesKeyValueStore
from ..framework.utilities.secrets import PanoptesSecretsStore
from ..framework.enrichment import PanoptesEnrichmentMultiGroupSet, \
    PanoptesEnrichmentCacheKeyValueStore, PanoptesEnrichmentGroupSet
from .enrichment_plugin import PanoptesEnrichmentPlugin

panoptes_context = None
panoptes_enrichment_task_context = None
celery = None


class PanoptesEnrichmentPluginAgentKeyValueStore(PanoptesKeyValueStore):
    """
    A custom Key/Value store for the Enrichment Plugin Agent which store data like a plugin's last execution time and \
    the last result production time
    """

    def __init__(self, panoptes_context):
        super(PanoptesEnrichmentPluginAgentKeyValueStore, self).__init__(
            panoptes_context, const.ENRICHMENT_PLUGIN_AGENT_KEY_VALUE_NAMESPACE)

        self._redis = panoptes_context.get_redis_connection(const.ENRICHMENT_REDIS_GROUP)


class PanoptesEnrichmentPluginKeyValueStore(PanoptesKeyValueStore):
    """
    A custom Key/Value store for the Panoptes Enrichment plugins

    Plugins executed by the Enrichment Plugin Agent get access to this store and can get/set values they need for their\
    operation
    """

    def __init__(self, panoptes_context):
        super(PanoptesEnrichmentPluginKeyValueStore, self).__init__(panoptes_context, const.PLUGINS_KEY_VALUE_NAMESPACE)


class PanoptesEnrichmentAgentContext(PanoptesContext):
    """
    This class implements a PanoptesContext with no clients - only used for setting up the Celery app
    """

    def __init__(self):
        super(PanoptesEnrichmentAgentContext, self).__init__(key_value_store_class_list=[],
                                                             create_message_producer=False,
                                                             create_zookeeper_client=False)


class PanoptesEnrichmentTaskContext(PanoptesContext):
    """
    This class implements a PanoptesContext with the following:
        - A Key/Value Store for Enrichment Plugin Agent
        - A Key/Value Store for Panoptes Plugins
        - A Asynchronous Message Producer
        - A Zookeeper Client
    """

    def __init__(self):
        super(PanoptesEnrichmentTaskContext, self).__init__(
                key_value_store_class_list=[PanoptesEnrichmentPluginAgentKeyValueStore,
                                            PanoptesEnrichmentPluginKeyValueStore,
                                            PanoptesEnrichmentCacheKeyValueStore,
                                            PanoptesSecretsStore,
                                            PanoptesResourcesKeyValueStore],
                create_message_producer=False, async_message_producer=False, create_zookeeper_client=True)


class PanoptesEnrichmentPluginAgentError(PanoptesBaseException):
    """
    The exception class for Enrichment Plugin Agent runtime errors
    """
    pass


@task
def enrichment_plugin_task(enrichment_plugin_name, resource_key):
    """
    The main method of the Enrichment Plugin Agent

    This method, called by Celery, loads and executes the specified plugin. The workflow is as follows:
        - Locate and load plugin through `yapsy's <http://yapsy.sourceforge.net/>`_ plugin manager
        - Evaluate if the plugin should be executed right now
        - If yes, then create a plugin context
        - Attempt to get a lock for the instance of the plugin (unique combination of the plugin name and configuration)
        - If the lock is acquired, then attempt to execute the plugin
        - If the plugin executes successfully, update the timestamp in the Enrichment Plugin Agent Key/Value store
        - If the plugin produced a non-zero result set, send the resource over the message bus
        - If sending over the message succeeds, update the timestamp in the Enrichment Plugin Agent Key/Value store

    Args:
        enrichment_plugin_name (str): The name of the plugin to be executed
        resource_key (dict): The dictionary that contains the resource key of the resource this enrichment plugin
        should collect info

    Returns:
        None

    """
    global panoptes_enrichment_task_context

    if panoptes_enrichment_task_context is None:
        try:
            panoptes_enrichment_task_context = PanoptesEnrichmentTaskContext()
        except Exception as e:
            sys.exit('Could not create a Panoptes Enrichment Task Context: %s' % (repr(e)))

    logger = panoptes_enrichment_task_context.logger

    try:
        resource = PanoptesResourceStore(panoptes_enrichment_task_context).get_resource(resource_key)
        plugin_runner = PanoptesPluginWithEnrichmentRunner(
            plugin_name=enrichment_plugin_name,
            plugin_type='enrichment',
            plugin_class=PanoptesEnrichmentPlugin,
            plugin_info_class=PanoptesPluginInfo,
            plugin_data=resource,
            panoptes_context=panoptes_enrichment_task_context,
            plugin_agent_kv_store_class=PanoptesEnrichmentPluginAgentKeyValueStore,
            plugin_kv_store_class=PanoptesEnrichmentPluginKeyValueStore,
            plugin_secrets_store_class=PanoptesSecretsStore,
            plugin_logger_name=const.ENRICHMENT_PLUGIN_AGENT_PLUGINS_CHILD_LOGGER_NAME,
            plugin_result_class=(PanoptesEnrichmentGroupSet, PanoptesEnrichmentMultiGroupSet),
            results_callback=_store_enrichment_data)
        plugin_runner.execute_plugin()
    except Exception as e:
        logger.error('[%s] Error executing plugin: %s' % (enrichment_plugin_name, str(e)))


def _store_enrichment_data(context, results, plugin):
    """
    Populate enrichment info on Redis kv store

    Args:
        context (PanoptesContext): The PanoptesContext being used by the Plugin Agent
        results (object): Enrichment data PanoptesEnrichmentGroupSet/PanoptesEnrichmentMultiGroupSet
        returned by the plugin
        plugin (PanoptesPluginInfo): The plugin object that produced this set of results

    Returns:
        None
    """
    logger = context.logger
    enrichment_key_value_store = context.get_kv_store(PanoptesEnrichmentCacheKeyValueStore)

    if isinstance(results, PanoptesEnrichmentGroupSet):
        resource = results.resource
        enrichment_groups = results.enrichment
        for enrichment_group in enrichment_groups:
            _update_enrichment_kv_store(logger, enrichment_key_value_store, resource, enrichment_group)
    elif isinstance(results, PanoptesEnrichmentMultiGroupSet):
        enrichment_group_sets = results.enrichment_group_sets
        for group_set in enrichment_group_sets:
            resource = group_set.resource
            for enrichment_group in group_set.enrichment:
                _update_enrichment_kv_store(logger, enrichment_key_value_store, resource, enrichment_group)


def _update_enrichment_kv_store(logger, enrichment_key_value_store, resource, enrichment_group):

    key = resource.resource_id + const.KV_NAMESPACE_DELIMITER + enrichment_group.namespace
    value = enrichment_group.serialize()

    logger.debug('Going to store enrichment info for resource id {} namespace {}'
                 .format(resource.resource_id, enrichment_group.namespace))

    try:
        enrichment_key_value_store.set(key, value, expire=enrichment_group.enrichment_ttl)
        logger.debug('Successfully populated enrichment info for resource id {} namespace {}'
                     .format(resource.resource_id, enrichment_group.namespace))
    except Exception as e:
        logger.error('Failed to store enrichment info for resource id {} namespace {}: {}'
                     .format(resource.resource_id, enrichment_group.namespace, e))


def start_enrichment_plugin_agent():
    """
    The entry point for the Enrichment Plugin Agent

    This method creates a Panoptes Context and the Celery Instance for the Enrichment Plugin Agent

    Returns:
        None
    """
    global panoptes_context, celery

    try:
        panoptes_context = PanoptesEnrichmentAgentContext()
    except Exception as e:
        sys.exit('Could not create a Panoptes Context: %s' % (str(e)))

    logger = panoptes_context.logger
    logger.info('Attempting to start Celery application')

    celery_config = PanoptesCeleryConfig(const.ENRICHMENT_PLUGIN_AGENT_CELERY_APP_NAME)

    try:
        celery = PanoptesCeleryInstance(panoptes_context, celery_config).celery
    except Exception as exp:
        sys.exit('Could not instantiate Celery application: %s' % str(exp))
    else:
        logger.info('Started Celery application: %s' % celery)


if get_calling_module_name() == const.CELERY_LOADER_MODULE:
    faulthandler.enable()
    start_enrichment_plugin_agent()



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.
"""

import time
import unittest
import logging
import os

from yahoo_panoptes.framework.validators import PanoptesValidators


class TestValidators(unittest.TestCase):
    def test_valid_nonempty_iterable_of_strings(self):
        class SampleIterator(object):
            def __init__(self, item):
                self.test_list = list()
                self.test_list.append(item)

            def __iter__(self):
                return iter(self.test_list)

            def __len__(self):
                return len(self.test_list)

        self.assertTrue(PanoptesValidators.valid_nonempty_iterable_of_strings(list('x')))
        self.assertTrue(PanoptesValidators.valid_nonempty_iterable_of_strings(set('x')))
        self.assertTrue(PanoptesValidators.valid_nonempty_iterable_of_strings(SampleIterator('x')))
        self.assertFalse(PanoptesValidators.valid_nonempty_iterable_of_strings('x'))
        self.assertFalse(PanoptesValidators.valid_nonempty_iterable_of_strings([]))
        self.assertFalse(PanoptesValidators.valid_nonempty_iterable_of_strings(['']))
        self.assertFalse(PanoptesValidators.valid_nonempty_iterable_of_strings(['x', '']))
        self.assertFalse(PanoptesValidators.valid_nonempty_iterable_of_strings(['x', set('x')]))

    def test_valid_timestamp(self):
        self.assertTrue(PanoptesValidators.valid_timestamp(timestamp=time.time()))
        self.assertTrue(PanoptesValidators.valid_timestamp(timestamp=time.time() - 86400))
        self.assertTrue(PanoptesValidators.valid_timestamp(timestamp=time.time() + 10))
        self.assertTrue(PanoptesValidators.valid_timestamp(timestamp=time.time() + 70, max_skew=120))
        self.assertFalse(PanoptesValidators.valid_timestamp(timestamp=0))
        self.assertFalse(PanoptesValidators.valid_timestamp(timestamp=time.time() + 120))
        self.assertFalse(PanoptesValidators.valid_timestamp(timestamp=time.time() - 604810))
        self.assertFalse(PanoptesValidators.valid_timestamp(timestamp=None))
        self.assertFalse(PanoptesValidators.valid_timestamp(timestamp=time.time(), max_age=None))
        self.assertFalse(PanoptesValidators.valid_timestamp(timestamp=time.time(), max_skew=None))

    def test_valid_plugin_result_class(self):
        self.assertFalse(PanoptesValidators.valid_plugin_result_class(None))
        self.assertTrue(PanoptesValidators.valid_plugin_result_class(str))

    def test_valid_logger(self):
        self.assertFalse(PanoptesValidators.valid_logger(None))
        self.assertTrue(PanoptesValidators.valid_logger(logging.Logger("test_logger")))

    def test_valid_callback(self):
        self.assertFalse(PanoptesValidators.valid_callback(object()))
        self.assertTrue(PanoptesValidators.valid_callback(object.__init__))

    def test_valid_hashable_object(self):
        self.assertTrue(PanoptesValidators.valid_hashable_object(object()))
        self.assertFalse(PanoptesValidators.valid_hashable_object(list()))

    def test_valid_port(self):
        self.assertTrue(PanoptesValidators.valid_port(1L))
        self.assertFalse(PanoptesValidators.valid_port(1.0))
        self.assertFalse(PanoptesValidators.valid_port(65540))
        self.assertFalse(PanoptesValidators.valid_port(0))
        self.assertFalse(PanoptesValidators.valid_port(-1))
        self.assertFalse(PanoptesValidators.valid_port("0"))
        self.assertFalse(PanoptesValidators.valid_port("1"))
        self.assertFalse(PanoptesValidators.valid_port(True))
        self.assertFalse(PanoptesValidators.valid_port(False))

    def test_valid_number(self):
        self.assertFalse(PanoptesValidators.valid_number(False))
        self.assertFalse(PanoptesValidators.valid_number("1.234"))
        self.assertTrue(PanoptesValidators.valid_number(2 ** (2 ** (2 ** (2 ** 2)))))
        self.assertTrue(PanoptesValidators.valid_number(1J))

    def test_valid_nonzero_integer(self):
        self.assertFalse(PanoptesValidators.valid_nonzero_integer(-1))
        self.assertFalse(PanoptesValidators.valid_nonzero_integer(0))
        self.assertFalse(PanoptesValidators.valid_nonzero_integer(-0L))
        self.assertFalse(PanoptesValidators.valid_nonzero_integer(1.0))
        self.assertFalse(PanoptesValidators.valid_nonzero_integer(""))
        self.assertFalse(PanoptesValidators.valid_nonzero_integer(False))
        self.assertFalse(PanoptesValidators.valid_nonzero_integer(True))
        self.assertTrue(PanoptesValidators.valid_nonzero_integer(2 ** (2 ** (2 ** (2 ** 2)))))

    def test_valid_positive_integer(self):
        self.assertTrue(PanoptesValidators.valid_positive_integer(0L))
        self.assertTrue(PanoptesValidators.valid_positive_integer(2 ** (2 ** (2 ** (2 ** 2)))))
        self.assertFalse(PanoptesValidators.valid_positive_integer(False))
        self.assertFalse(PanoptesValidators.valid_positive_integer(True))
        self.assertFalse(PanoptesValidators.valid_positive_integer(""))
        self.assertFalse(PanoptesValidators.valid_positive_integer(1.0))
        self.assertFalse(PanoptesValidators.valid_positive_integer(-5))
        self.assertTrue(PanoptesValidators.valid_positive_integer(-0L))

    def test_valid_nonempty_string(self):
        self.assertFalse(PanoptesValidators.valid_nonempty_string(""))
        self.assertTrue(PanoptesValidators.valid_nonempty_string("hello world"))
        self.assertFalse(PanoptesValidators.valid_nonempty_string(0))
        
    def test_valid_readable_path(self):
        self.assertTrue(PanoptesValidators.valid_readable_path(os.getcwd()))
        self.assertFalse(PanoptesValidators.valid_readable_path(os.getcwd() + '/test_dir'))
        self.assertFalse(PanoptesValidators.valid_readable_path('not a path'))
        self.assertFalse(PanoptesValidators.valid_readable_path(None))

    def test_valid_readable_file(self):
        self.assertTrue(PanoptesValidators.valid_readable_file(os.path.realpath(__file__)))
        self.assertFalse(PanoptesValidators.valid_readable_file(None))

    def test_valid_numeric_snmp_oid(self):
        self.assertTrue(PanoptesValidators.valid_numeric_snmp_oid(".1.3.6.1.4.1.2636.3.1.13.1.6.2.1.0.0"))
        self.assertFalse(PanoptesValidators.valid_numeric_snmp_oid("1.3.6.1.4.1.2636.3.1.13.1.6.2.1.0.0"))



import os
import unittest
from mock import Mock, patch

from tests.plugins.helpers import SNMPPollingPluginTestFramework, setup_module_default, tear_down_module_default
from yahoo_panoptes.plugins.polling.interface.plugin_polling_device_interface_metrics import \
    PluginPollingDeviceInterfaceMetrics, _MISSING_METRIC_VALUE, _INTERFACE_STATES

_MOCK_INTERFACE_ENTRY = '0'
pwd = os.path.dirname(os.path.abspath(__file__))


def setUpModule():
    return setup_module_default(plugin_pwd=pwd)


def tearDownModule():
    return tear_down_module_default()


class TestPluginPollingDeviceInterfaceMetrics(SNMPPollingPluginTestFramework, unittest.TestCase):
    plugin_conf = {
        'Core': {
            'name': 'Test Plugin',
            'module': 'test_plugin'
        },
        'main': {
            'execute_frequency': '60',
            'resource_filter': 'resource_class = "network"'
        },
        'enrichment': {
            'preload': 'self:interface'
        }
    }

    plugin_class = PluginPollingDeviceInterfaceMetrics
    path = pwd

    def test_missing_interface(self):
        plugin = self.plugin_class()
        plugin.run(self._plugin_context)
        self.assertEqual(plugin.get_bits_in(_MOCK_INTERFACE_ENTRY), _MISSING_METRIC_VALUE)
        self.assertEqual(plugin.get_bits_out(_MOCK_INTERFACE_ENTRY), _MISSING_METRIC_VALUE)
        self.assertEqual(plugin.get_total_packets_in(_MOCK_INTERFACE_ENTRY), _MISSING_METRIC_VALUE)
        self.assertEqual(plugin.get_total_packets_out(_MOCK_INTERFACE_ENTRY), _MISSING_METRIC_VALUE)

    def test_get_state_val(self):
        assert self.plugin_class._get_state_val('2') == _INTERFACE_STATES.DOWN
        assert self.plugin_class._get_state_val('1') == _INTERFACE_STATES.UP
        assert self.plugin_class._get_state_val('1234') == _INTERFACE_STATES.UNKNOWN

    def test_metric_not_number(self):
        mock_is_instance = Mock(return_value=False)
        plugin = self.plugin_class()
        with patch('yahoo_panoptes.plugins.polling.interface.plugin_polling_device_interface_metrics.'
                   'isinstance', mock_is_instance):
            results = plugin.run(self._plugin_context)
            self.assertEqual(len(results), 2)

    def test_iftable_stats_exception_handling(self):
        mock_get_type = Mock(side_effect=Exception)
        plugin = self.plugin_class()
        plugin.get_type = mock_get_type
        self.assertRaises(Exception, plugin.run(self._plugin_context))

    def test_ifxtable_stats_exception_handling(self):
        mock_get_bits_in = Mock(side_effect=Exception)
        plugin = self.plugin_class()
        plugin.get_bits_in = mock_get_bits_in
        self.assertRaises(Exception, plugin.run(self._plugin_context))

    def test_dot3table_stats_exception_handling(self):
        mock_get_errors_frame = Mock(side_effect=Exception)
        plugin = self.plugin_class()
        plugin.get_errors_frame = mock_get_errors_frame
        self.assertRaises(Exception, plugin.run(self._plugin_context))



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

yahoo_panoptes.utilities namespace package
"""
__import__('pkg_resources').declare_namespace(__name__)



from yahoo_panoptes.plugins.enrichment.interface.plugin_enrichment_interface import PluginEnrichmentInterface


class PluginEnrichmentAristaInterface(PluginEnrichmentInterface):
    """
    InterfaceEnrichment class for Arista devices.
    """
    def get_parent_interface_name(self, index):
        """
        Gets the parent interface name for the interface associated with the provided index

        Args:
            index (int): The index used to look up the associated interface in self._interface_table

        Returns:
            string: The name of the parent interface, or self._MISSING_VALUE_STRING if the interface has no parent.
                    For Arista devices, this is everything to the left of the '/' in the interface name, if a '/' is
                    present.
        """
        interface_name = self.get_interface_name(index)
        if '/' in interface_name:
            return interface_name.split('/')[0]
        return self._MISSING_VALUE_STRING

    def get_parent_interface_configured_speed(self, index):
        parent_name = self.get_parent_interface_name(index)
        if parent_name is not self._MISSING_VALUE_STRING:
            return 4 * self.get_configured_speed(index)
        return self._MISSING_METRIC_VALUE

    def get_parent_interface_media_type(self, index):
        return self.get_media_type(index)

    def get_parent_interface_port_speed(self, index):
        return self.get_parent_interface_configured_speed(index)



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.
"""

from mock import patch
from time import sleep
import threading

from kazoo.client import KazooState
from kazoo.testing import KazooTestCase

from yahoo_panoptes.framework.context import PanoptesContext
from yahoo_panoptes.framework.utilities.lock import PanoptesLock

kazoo_client = None


class TestPanoptesLock(KazooTestCase):
    @staticmethod
    def make_event():
        return threading.Event()

    @staticmethod
    def mock_kazoo_client():
        return kazoo_client

    @patch.object(PanoptesContext, '_get_zookeeper_client', mock_kazoo_client)
    def test_panoptes_lock(self):
        global kazoo_client
        kazoo_client = self.client

        connected = threading.Event()
        lost_connection = threading.Event()

        def _listener(state):
            if state == KazooState.CONNECTED:
                connected.set()
            else:
                lost_connection.set()

        self.client.add_listener(_listener)

        panoptes_context = PanoptesContext(config_file='tests/config_files/test_panoptes_config.ini',
                                           create_zookeeper_client=True)

        # Test that bad parameters fail
        with self.assertRaises(AssertionError):
            PanoptesLock(context=None, path='/lock', timeout=5, retries=0, identifier='test')
        with self.assertRaises(AssertionError):
            PanoptesLock(context=panoptes_context, path='/lock', timeout=5, retries=None, identifier='test')
        with self.assertRaises(AssertionError):
            PanoptesLock(context=panoptes_context, path='/lock', timeout=5, retries=-1, identifier='test')
        with self.assertRaises(AssertionError):
            PanoptesLock(context=panoptes_context, path='/lock', timeout=5, retries='1', identifier='test')
        with self.assertRaises(AssertionError):
            PanoptesLock(context=panoptes_context, path=None, timeout=5, retries=0, identifier='test')
        with self.assertRaises(AssertionError):
            PanoptesLock(context=panoptes_context, path='lock', timeout=5, retries=0, identifier='test')
        with self.assertRaises(AssertionError):
            PanoptesLock(context=panoptes_context, path='/lock', timeout=None, retries=0, identifier='test')
        with self.assertRaises(AssertionError):
            PanoptesLock(context=panoptes_context, path='/lock', timeout=-1, retries=0, identifier='test')
        with self.assertRaises(AssertionError):
            PanoptesLock(context=panoptes_context, path='/lock', timeout=-1, retries=0, identifier='test')
        with self.assertRaises(AssertionError):
            PanoptesLock(context=panoptes_context, path='/lock', timeout=5, retries=0, identifier=None)
        with self.assertRaises(AssertionError):
            PanoptesLock(context=panoptes_context, path='/lock', timeout=5, retries=0, identifier=1)

        # Acquire lock with unlimited retries
        lock = PanoptesLock(context=panoptes_context, path='/lock', timeout=5, retries=0, identifier='test')
        self.assertEquals(lock.locked, True)

        # Release the lock
        lock.release()
        self.assertEquals(lock.locked, False)

        # Acquire lock with only one retry
        lock1 = PanoptesLock(context=panoptes_context, path='/lock', timeout=5, identifier='test')
        self.assertEquals(lock1.locked, True)

        # Try an acquire an acquired lock - this should fail
        lock2 = PanoptesLock(context=panoptes_context, path='/lock', timeout=5, identifier='test')
        self.assertEquals(lock2.locked, False)
        lock1.release()
        self.assertEquals(lock1.locked, False)
        lock2.release()

        # Acquire the lock, lose connection and lose the lock and acquire it again on reconnection
        lock = PanoptesLock(context=panoptes_context, path='/lock', timeout=5, identifier='test')
        self.assertEquals(lock.locked, True)
        self.lose_connection(self.make_event)
        # Block till the client disconnects - or 30 seconds pass
        lost_connection.wait(30)
        # Verify that the client actually lost the connection
        self.assertEquals(lost_connection.is_set(), True)
        # Give it time to cleanup the lock
        # TODO: There is a timing issue here - if we sleep too long before checking the state of the lock, we
        # might get reconnected and reacquire the lock, which is why we check if connected.is_set is NOT set before the
        # assert
        sleep(0.1)
        # The lock should be not be set after we loose a connection
        if not connected.is_set():
            self.assertEquals(lock.locked, False)
        # Block till the client reconnects - or 30 seconds pass
        connected.wait(30)
        # Verify that the client actually reconnected
        self.assertEquals(connected.is_set(), True)
        # Give it time to reacquire the lock
        sleep(3)
        self.assertEquals(lock.locked, True)



"""
Juniper device snmp mib classes
"""
from .base import oid, SNMPGauge32, SNMPInteger, SNMPInteger32, SNMPString
from .snmpv2 import MibSNMPV2


class MibJuniper(MibSNMPV2):
    """
    Generic Juniper MIB
    """
    juniperMIB = oid('.1.3.6.1.4.1.2636')
    jnxMibs = juniperMIB + oid('3')                                                    # 1.3.6.1.4.1.2636.3
    jnxBoxAnatomy = jnxMibs + oid('1')                                                 # 1.3.6.1.4.1.2636.3.1
    jnxOperatingTable = jnxBoxAnatomy + oid('13')                                      # 1.3.6.1.4.1.2636.3.1.13
    jnxOperatingEntry = jnxOperatingTable + oid('1')                                   # 1.3.6.1.4.1.2636.3.1.13.1
    jnxOperatingContentsIndex = jnxOperatingEntry + oid('1', SNMPInteger32)            # 1.3.6.1.4.1.2636.3.1.13.1.1
    jnxOperatingL1Index = jnxOperatingEntry + oid('2', SNMPInteger)                    # 1.3.6.1.4.1.2636.3.1.13.1.2
    jnxOperatingDescr = jnxOperatingEntry + oid('5', SNMPString)                       # 1.3.6.1.4.1.2636.3.1.13.1.5
    jnxOperatingState = jnxOperatingEntry + oid('6', SNMPInteger)                      # 1.3.6.1.4.1.2636.3.1.13.1.6
    jnxOperatingTemp = jnxOperatingEntry + oid('7', SNMPGauge32)                       # 1.3.6.1.4.1.2636.3.1.13.1.7
    jnxOperatingCPU = jnxOperatingEntry + oid('8', SNMPGauge32)                        # 1.3.6.1.4.1.2636.3.1.13.1.8
    jnxOperatingBuffer = jnxOperatingEntry + oid('11', SNMPGauge32)                    # 1.3.6.1.4.1.2636.3.1.13.1.11
    jnxOperatingHeap = jnxOperatingEntry + oid('12', SNMPGauge32)                      # 1.3.6.1.4.1.2636.3.1.13.1.12
    jnxOperatingMemory = jnxOperatingEntry + oid('15', SNMPGauge32)                    # 1.3.6.1.4.1.2636.3.1.13.1.15
    jnxOperating1MinLoadAvg = jnxOperatingEntry + oid('20', SNMPGauge32)               # 1.3.6.1.4.1.2636.3.1.13.1.20
    jnxOperating5MinLoadAvg = jnxOperatingEntry + oid('21', SNMPGauge32)               # 1.3.6.1.4.1.2636.3.1.13.1.21
    jnxOperating15MinLoadAvg = jnxOperatingEntry + oid('22', SNMPGauge32)              # 1.3.6.1.4.1.2636.3.1.13.1.22

    jnxJsSPUMonitoringCPUUsage = juniperMIB + oid('3.39.1.12')  # TODO no entry for this oid entries for ...39.1.11



import os

import unittest

from yahoo_panoptes.plugins.enrichment.interface.juniper.plugin_enrichment_interface_juniper import \
    PluginEnrichmentJuniperInterface
from tests.plugins.helpers import setup_module_default, tear_down_module_default, SNMPEnrichmentPluginTestFramework

pwd = os.path.dirname(os.path.abspath(__file__))

def setUpModule():
    return setup_module_default(plugin_pwd=pwd)


def tearDownModule():
    return tear_down_module_default()


class TestJuniperQFX5200InterfaceEnrichmentPlugin(SNMPEnrichmentPluginTestFramework, unittest.TestCase):
    resource_id = 'router1'
    resource_plugin = 'router_discovery_plugin'
    resource_site = 'dc1'
    resource_class = 'network'
    resource_subclass = 'router'
    resource_type = 'juniper'

    plugin_class = PluginEnrichmentJuniperInterface
    path = pwd

    def setUp(self):
        super(TestJuniperQFX5200InterfaceEnrichmentPlugin, self).setUp()
        self._panoptes_resource.resource_metadata['model'] = 'QFX5200'



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.
"""

import os
import unittest

from mock import patch

from yahoo_panoptes.framework.context import PanoptesContext
from yahoo_panoptes.framework.plugins.manager import PanoptesPluginManager
from yahoo_panoptes.polling.polling_plugin import PanoptesPollingPlugin

from test_framework import panoptes_mock_redis_strict_client, panoptes_mock_kazoo_client


class TestPanoptesPluginManagerContext(PanoptesContext):
    def __init__(self):
        my_dir = os.path.dirname(os.path.realpath(__file__))
        panoptes_test_conf_file = os.path.join(my_dir, 'config_files/test_panoptes_config.ini')
        super(TestPanoptesPluginManagerContext, self).__init__(config_file=panoptes_test_conf_file,
                                                               create_zookeeper_client=True)


class TestPanoptesPluginManager(unittest.TestCase):
    @patch('redis.StrictRedis', panoptes_mock_redis_strict_client)
    @patch('kazoo.client.KazooClient', panoptes_mock_kazoo_client)
    def test_panoptes_plugin_manager(self):
        matching_test_plugins_found = 0

        context = TestPanoptesPluginManagerContext()
        plugin_manager = PanoptesPluginManager(plugin_type='polling',
                                               plugin_class=PanoptesPollingPlugin,
                                               panoptes_context=context)

        plugins = plugin_manager.getPluginsOfCategory('polling')

        self.assertGreater(len(plugins), 0)

        for plugin in plugins:
            if 'Test Polling Plugin' in plugin.name:
                matching_test_plugins_found += 1

        self.assertEqual(matching_test_plugins_found, 2)

        polling_plugin = plugin_manager.getPluginByName(name='Test Polling Plugin', category='polling')

        self.assertEqual(polling_plugin.name, 'Test Polling Plugin')
        self.assertEqual(os.path.split(polling_plugin.config_filename)[1], 'plugin_polling_test.panoptes-plugin')

        polling_plugin_second_instance = plugin_manager.getPluginByName(name='Test Polling Plugin Second Instance',
                                                                        category='polling')

        self.assertEqual(polling_plugin_second_instance.name, 'Test Polling Plugin Second Instance')
        self.assertEqual(os.path.split(polling_plugin_second_instance.config_filename)[1],
                         'plugin_polling_test_second_instance.panoptes-plugin')



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This module declares an abstract base class for Panoptes Polling Plugins - and related helpers

A Polling Plugin's role is to return a collection of metrics that have been monitored by the system
"""
import abc

import six

from ..framework.exceptions import PanoptesBaseException
from ..framework.metrics import PanoptesMetricSet
from ..framework.plugins.context import PanoptesPluginContext
from ..framework.plugins.panoptes_base_plugin import PanoptesBasePlugin, PanoptesPluginInfo, \
    PanoptesPluginConfigurationError


class PanoptesPollingPluginError(PanoptesBaseException):
    """
    The class for Polling Plugin runtime errors
    """
    pass


class PanoptesPollingPluginConfigurationError(PanoptesPluginConfigurationError):
    """
    The class for Polling Plugin configuration errors
    """
    pass


class PanoptesPollingPluginInfo(PanoptesPluginInfo):
    """
    This class extends the PanoptesPluginInfo to contain the resource_filter associated with a Polling Plugin

    Args:
        plugin_name(str): The name of the plugin
        plugin_path(str): The 'path' of the plugin - either the absolute path of the directory that contains the plugin\
        module or the path of the module file without the '.py' extension
    """
    def __init__(self, plugin_name, plugin_path):
        super(PanoptesPollingPluginInfo, self).__init__(plugin_name, plugin_path)

    @property
    def resource_filter(self):
        """
        The Resource DSL query to be applied to resources passed to this plugin

        Returns:
            str: The Resource DSL filter specified in the plugin configuration file

        Raises:
            PanoptesPollingPluginConfigurationError: This error if raised if no Resource DSL filter is specified in the\
            plugin configuration file
        """
        try:
            return self.details.get('main', 'resource_filter')
        except Exception as e:
            raise PanoptesPollingPluginConfigurationError(
                'No resource filter specified or error in parsing resource filter for plugin %s: %s' % (
                    self.name, str(e)))


@six.add_metaclass(abc.ABCMeta)
class PanoptesPollingPlugin(PanoptesBasePlugin):
    """
    The base class for all Panoptes Polling plugins

    A Polling Plugin's role is to return a collection of metrics associated with a resource
    """

    @abc.abstractmethod
    def run(self, context):
        """
        The entry point for Polling plugins

        Args:
            context (PanoptesPluginContext): The context of the plugin contains the plugins configuration, the logger \
            it should use and a client for the system wide plugins KV store

        Returns:
            PanoptesMetricSet: A collection of metrics polled by the plugin

       Raises:
            PanoptesPluginConfigurationError: If the plugin cannot proceed with the configuration provided to it \
            through the context, it should raise a PanoptesPluginConfigurationError

        """
        pass



from mockredis import MockRedis
from redis.exceptions import ConnectionError, TimeoutError


class PanoptesMockRedis(MockRedis):
    def __init__(self, bad_connection=False, timeout=False, **kwargs):
        if bad_connection:
            raise ConnectionError
        super(PanoptesMockRedis, self).__init__(**kwargs)
        self.connection_pool = 'mockredis connection pool'
        self.timeout = timeout

    def get(self, key):
        if self.timeout:
            raise TimeoutError
        else:
            return super(PanoptesMockRedis, self).get(key)

    def set(self, key, value, ex=None, px=None, nx=False, xx=False):
        if self.timeout:
            raise TimeoutError
        else:
            return super(PanoptesMockRedis, self).set(key, value, ex=ex, px=px, nx=nx, xx=xx)



#!/usr/bin/env python
import sys
import os
from setuptools import setup, find_packages

version = '1.1'

if ('CI' in os.environ) or ('CONTINUOUS_INTEGRATION' in os.environ):
    if 'SCREWDRIVER' in os.environ:
        build_number = os.environ['BUILD_NUMBER']
    elif 'TRAVIS' in os.environ:
        build_number = os.environ['TRAVIS_BUILD_NUMBER']
    else:
        sys.exit('We currently only support building CI builds with Screwdriver or Travis')

    IN_CI_PIPELINE = True
    version = '.'.join([version, build_number])
else:
    version += '.0'

# Read the long description from README.md
with open('README.md') as f:
    long_description = f.read()


def package_scripts():
    """
    Update the "scripts" parameter of the setup_arguments with any scripts
    found in the "scripts" directory.
    :return:
    """
    scripts_list = []
    if os.path.isdir('scripts'):
        scripts_list = [
            os.path.join('scripts', f) for f in os.listdir('scripts')
            ]
    return scripts_list


setup_arguments = {
    'version': version,
    'long_description': long_description,
    'long_description_content_type': "text/markdown",
    'packages': find_packages(),
    'package_data': {
        'yahoo_panoptes.framework': [
            'panoptes_configspec.ini',
        ],
        'yahoo_panoptes.consumers.influxdb': [
            'influxdb_consumer_configspec.ini',
        ]
    },
    'scripts': package_scripts(),
}
args = setup_arguments


if __name__ == '__main__':
    setup(**setup_arguments)



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This module declares an abstract base class for Panoptes Discovery Plugins - and related helpers

A Discovery Plugin's role is to return a collection of resources that should be monitored by the system
"""
import abc

import six

from ..framework.exceptions import PanoptesBaseException
from ..framework.plugins.context import PanoptesPluginContext
from ..framework.plugins.panoptes_base_plugin import PanoptesBasePlugin


class PanoptesDiscoveryPluginError(PanoptesBaseException):
    """
    The class for Discovery Plugin runtime errors
    """
    pass


@six.add_metaclass(abc.ABCMeta)
class PanoptesDiscoveryPlugin(PanoptesBasePlugin):
    """
    The base class for all Panoptes Discovery plugins

    A Discovery Plugin's role is to return a collection of resources that should be monitored by the system
    """

    @abc.abstractmethod
    def run(self, context):
        """
        The entry point for Discovery Manager plugins

        Args:
            context (PanoptesPluginContext): The context of the plugin contains the plugins configuration, the logger \
            it should use and a client for the system wide plugins KV store

        Returns:
            PanoptesResourceSet: A collection of resources discovered by the plugin

       Raises:
            PanoptesPluginConfigurationError: If the plugin cannot proceed with the configuration provided to it \
            through the context, it should raise a PanoptesPluginConfigurationError

        """
        pass



from __future__ import unicode_literals

from ...exceptions import PanoptesBaseException


class PanoptesSNMPException(PanoptesBaseException):
    """The base Easy SNMP exception which covers all exceptions raised."""
    pass


class PanoptesSNMPConnectionException(PanoptesSNMPException):
    """Indicates a problem connecting to the remote host."""
    pass


class PanoptesSNMPTimeoutException(PanoptesSNMPConnectionException):
    """Raised when an SNMP request times out."""
    pass


class PanoptesSNMPUnknownObjectIDException(PanoptesSNMPException):
    """Raised when an inexisted OID is requested."""
    pass


class PanoptesSNMPNoSuchNameException(PanoptesSNMPException):
    """
    Raised when an OID is requested which may be an invalid object name
    or invalid instance (only applies to SNMPv1).
    """
    pass


class PanoptesSNMPNoSuchObjectException(PanoptesSNMPException):
    """
    Raised when an OID is requested which may have some form of existence but
    an invalid object name.
    """
    pass


class PanoptesSNMPNoSuchInstanceException(PanoptesSNMPException):
    """
    Raised when a particular OID index requested from Net-SNMP doesn't exist.
    """
    pass


class PanoptesSNMPUndeterminedTypeException(PanoptesSNMPException):
    """
    Raised when the type cannot be determine when setting the value of an OID.
    """
    pass



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This module defines classes related to all Panoptes Plugins
"""
import abc
import hashlib
import json
import os
import time
from collections import defaultdict

import six
from cached_property import threaded_cached_property
from yapsy.IPlugin import IPlugin
from yapsy.PluginInfo import PluginInfo

from .. import const
from ..exceptions import PanoptesBaseException
from ..validators import PanoptesValidators
from ..utilities.helpers import get_client_id, get_module_mtime, normalize_plugin_name
from ..utilities.lock import PanoptesLock
from .context import PanoptesPluginContext


class PanoptesPluginConfigurationError(PanoptesBaseException):
    """
    The exception class for Panoptes Plugin configuration errors
    """
    pass


class PanoptesPluginRuntimeError(PanoptesBaseException):
    """
    The exception class for Panoptes Plugin runtime errors
    """
    pass


class PanoptesPluginInfoValidators(object):
    @classmethod
    def valid_plugin_info_class(cls, plugin_info_class):
        """
        valid_plugin_info_class(cls, plugin_info_class)

        Checks if the passed class is a subclass of PanoptesPluginInfo

        Args:
            plugin_info_class (cls): The class to check

        Returns:
            bool: True if the class is not null and is an subclass of PanoptesPluginInfo
        """
        return plugin_info_class and issubclass(plugin_info_class, PanoptesPluginInfo)


class PanoptesPluginInfo(PluginInfo):
    """
    This class represents all the metadata associated with a plugin such as:
        - Configuration read from the plugin definition file
        - The mtime of the plugin module and the configuration
        - The last executed time and age
        - The last results time and age
        - The Panoptes Context associated with the Plugin Agent that is executing this plugin
        - The Key/Value store associated that the plugin can use
        - Whether this plugin should be executed right now
        - The 'signature' of the plugin - which is a hash of it's config and data
        - The data to be passed to the plugin
        - A lock that the plugin can acquire

    Args:
        plugin_name(str): The name of the plugin
        plugin_path(str): The 'path' of the plugin - either the absolute path of the directory that contains the plugin
        module or the path of the module file without the '.py' extension
    """

    def __init__(self, plugin_name, plugin_path):
        super(PanoptesPluginInfo, self).__init__(plugin_name, plugin_path)
        self._normalized_name = normalize_plugin_name(plugin_name)
        self._panoptes_context = None
        self._kv_store_class = None
        self._last_executed = None
        self._last_executed_key = None
        self._last_results = None
        self._last_results_key = None
        self._config_filename = None
        self._data = None
        self._lock = None

    def __repr__(self):
        return 'PanoptesPluginInfo: ' \
               'Normalized name: {}, '\
               'Config file: {}, '\
               'Panoptes context: {}, '\
               'KV store class: {}, '\
               'Last executed timestamp: {}, '\
               'Last executed key: {}, '\
               'Last results timestamp: {}, '\
               'Last results key: {}, '\
               'Data: {}, '\
               'Lock: {}'.format(self.normalized_name,
                                 str(self.config_filename),
                                 repr(self.panoptes_context) if self._panoptes_context else 'None',
                                 str(self.kv_store_class.__name__),
                                 str(self.last_executed),
                                 str(self.last_executed_key),
                                 str(self.last_results),
                                 str(self.last_results_key),
                                 'Data object passed' if (self.data is not None) else 'None',
                                 'Lock is set' if (self.lock is not None) else 'None')

    def _get_key(self, suffix):
        """
        Returns the key that should be used to get the plugin's metadata from the Key/Value store

        The key returned contains the normalized name signature of the plugin (see below for their definition)

        Args:
            suffix(str): The suffix is the specific piece of metadata to lookup (e.g. 'last_results'

        Returns:
            str: The lookup key that should be used

        """
        return 'plugin_metadata:'\
               + self._normalized_name\
               + ':' + self.signature\
               + ':' + suffix

    @threaded_cached_property
    def normalized_name(self):
        """
        The normalized named of the plugin

        Normalized name represents a 'safe' name that can be used throughout the Panoptes system

        Returns:
            str: The normalized plugin name
        """
        return self._normalized_name

    @threaded_cached_property
    def normalized_category(self):
        """
        The normalized named of the plugin's category

        Normalized name represents a 'safe' name that can be used throughout the Panoptes system

        Returns:
            str: The normalized plugin name
        """
        return normalize_plugin_name(self.category)

    @property
    def moduleMtime(self):
        """
        The plugin module's mtime in Unix Epoch format

        Returns:
            int: The plugin module's mtime in Unix Epoch format

        Raises:
            PanoptesPluginConfigurationError: This exception is raised if the plugin module's mtime cannot be determined
        """
        try:
            return get_module_mtime(self.path)
        except Exception as e:
            raise PanoptesPluginConfigurationError(
                'Could not get mtime for module file/directory "%s" for plugin "%s": %s' % (
                    self.path, self.name, str(e)))

    @property
    def configMtime(self):
        """
        The plugin config's time in Unix Epoch format

        Returns:
            int: The plugin config's mtime in Unix Epoch format

        Raises:
            PanoptesPluginConfigurationError: This exception is raised if the plugin config's mtime cannot be determined
        """
        try:
            return int(os.path.getmtime(self._config_filename))
        except Exception as e:
            raise PanoptesPluginConfigurationError(
                'Could not get mtime for configuration file "%s" for plugin "%s": %s' % (
                    self._config_filename, self.name, str(e)))

    @property
    def execute_frequency(self):
        """
        The execution frequency of the plugin in seconds specified in the plugin's config file

        Returns zero in case the execution frequency is not specified or is not an integer

        Returns:
            int: The execution frequency of the plugin in seconds
        """
        try:
            return self.details.getint('main', 'execute_frequency')
        except:
            return 0

    @property
    def results_cache_age(self):
        """
        The results cache age of the plugin in seconds specified in the plugin's config file

        Returns zero in case the results cache age is not specified or is not an integer

        Returns:
            int: The results cache age of the plugin in seconds
        """
        try:
            return self.details.getint('main', 'results_cache_age')
        except:
            return 0

    @threaded_cached_property
    def config(self):
        """
        The configuration of the plugin

        This method folds the configparser which holds the plugin's parsed configuration file into a dictionary

        Returns:
            dict: The configuration of the plugin in sorted order.

        """
        config = defaultdict(dict)
        for section in self.details.sections():
            for option in self.details.options(section):
                config[section][option] = self.details.get(section, option)

        return config

    @property
    def panoptes_context(self):
        """
        The Panoptes Context associated with the Plugin Agent executing the plugin

        Returns:
            PanoptesContext: The Panoptes Context associated with the Plugin Agent executing the plugin

        Raises:
            PanoptesPluginConfigurationError: If no Panoptes Context is associated with the plugin
        """
        if not self._panoptes_context:
            raise PanoptesPluginConfigurationError('No Panoptes Context associated with plugin "%s"' % self.name)
        return self._panoptes_context

    @panoptes_context.setter
    def panoptes_context(self, panoptes_context):
        """
        Sets the Panoptes Context associated with this plugin instance

        Args:
            panoptes_context (PanoptesContext): The Panoptes Context for with the Plugin Agent executing the plugin

        Returns:
            None
        """
        self._panoptes_context = panoptes_context

    @property
    def kv_store_class(self):
        """
        The Key/Value store class that the plugin instance is using

        Returns:
            class: The Key/Value store class that the plugin instance is using

        """
        return self._kv_store_class

    @kv_store_class.setter
    def kv_store_class(self, kv_store_class):
        """
        Sets the Key/Value store class that the plugin instance should use

        Args:
            kv_store_class (PanoptesKeyValueStore): The Key/Value store class that the plugin instance should use

        Returns:
            None
        """
        self._kv_store_class = kv_store_class

    @property
    def metadata_kv_store(self):
        """
        The Key/Value store class that the plugin info object is using

        Returns:
            PanoptesKeyValueStore: The Key/Value store class that the plugin info object is using
        """
        return self._panoptes_context.get_kv_store(self.kv_store_class)

    @property
    def last_executed_key(self):
        """
        The key to lookup to get the last executed time for the plugin

        Returns:
            str: he key to lookup to get the last executed time for the plugin
        """
        return self._get_key('last_executed')

    @property
    def last_executed(self):
        """
        The last execution time of the plugin in Unix Epoch format

        Returns zero if the plugin has never been executed or there is an error in fetching the last execution time

        Returns:
            int: The last execution time of the plugin in Unix Epoch format
        """
        if self._last_executed is not None:
            return self._last_executed

        try:
            self._last_executed = int(self.metadata_kv_store.get(self.last_executed_key))
        except:
            self._last_executed = 0

        return self._last_executed

    @last_executed.setter
    def last_executed(self, timestamp):
        """
        Sets the last execution time of the plugin

        Args:
            timestamp (int): The last execution time of the plugin in Unix Epoch format

        Returns:
            None
        """
        assert PanoptesValidators.valid_positive_integer(timestamp), "timestamp must be a positive integer."
        try:
            self.metadata_kv_store.set(self.last_executed_key, str(timestamp),
                                       expire=const.PLUGIN_AGENT_PLUGIN_TIMESTAMPS_EXPIRE)
        except Exception as exp:
            self.panoptes_context.logger.error('Could not store value for last successful execution time for plugin '
                                               '"%s": %s' % (self.name, str(exp)))

    @property
    def last_executed_age(self):
        """
        The last execution age (in seconds) of the plugin

        Returns:
            int: The last execution age (in seconds) of the plugin
        """
        return int(time.time()) - self.last_executed

    @property
    def last_results_key(self):
        """
        The key to lookup to get the last results time for the plugin

        Returns:
            str: he key to lookup to get the last results time for the plugin
        """
        return self._get_key('last_results')

    @property
    def last_results(self):
        """
        The last results time of the plugin in Unix Epoch format

        Returns zero if the plugin has never been produced results or there is an error in fetching the last results
        time

        Returns:
            int: The last results time of the plugin in Unix Epoch format
        """
        if self._last_results is not None:
            return self._last_results

        try:
            self._last_results = int(self.metadata_kv_store.get(self.last_results_key))
        except:
            self._last_results = 0

        return self._last_results

    @last_results.setter
    def last_results(self, timestamp):
        """
        Sets the last results time of the plugin

        Args:
            timestamp (int): The last results time of the plugin in Unix Epoch format

        Returns:
            None
        """
        assert PanoptesValidators.valid_positive_integer(timestamp), "timestamp must be a positive integer."
        try:
            self.metadata_kv_store.set(self.last_results_key, str(timestamp),
                                       expire=const.PLUGIN_AGENT_PLUGIN_TIMESTAMPS_EXPIRE)
        except Exception as exp:
            self.panoptes_context.logger.error('Could not store value for last successful results time for plugin '
                                               '"%s": %s' % (self.name, str(exp)))

    @property
    def last_results_age(self):
        """
        The last results age (in seconds) of the plugin

        Returns:
            int: The last results age (in seconds) of the plugin
        """
        return int(time.time()) - self.last_results

    @property
    def execute_now(self):
        """
        Decides whether a plugin should be executed right now or not

        True if the plugin's last execution age is greater than the execution frequency OR it's module/config mtimes are
        greater than the last execution time

        The motivation is to ensure that a plugin is re-run as soon as possible after it's module or config have changed

        Similar check for plugin results

        Returns:
            bool: True if the plugin should be executed now - false otherwise
        """
        logger = self.panoptes_context.logger
        skew = self.panoptes_context.config_dict['main']['plugins_skew']

        if self.last_executed_age + skew < self.execute_frequency:
            if (self.last_executed > self.moduleMtime) and (
                        self.last_executed > self.configMtime):
                logger.info('Skipping execution of plugin "%s" since it was last executed at %s (UTC), which is less '
                            'then %s seconds ago while the plugin execution frequency is set to %s seconds'
                            % (self.name, self.last_executed, self.last_executed_age,
                               self.execute_frequency))
                return False

        if self.last_results_age + skew < self.results_cache_age:
            if (self.last_results > self.moduleMtime) and (
                        self.last_results > self.configMtime):
                logger.info('Skipping execution of plugin "%s" since it last produced a resource set at %s (UTC), '
                            'which is less then %s seconds ago while the results cache age is set to %s seconds'
                            % (self.name, self.last_results, self.last_results_age,
                               self.results_cache_age))
                return False

        return True

    @property
    def config_filename(self):
        """
        The name of configuration file associated with the plugin

        Returns:
            str: The name of configuration file associated with the plugin
        """
        return self._config_filename

    @config_filename.setter
    def config_filename(self, filename):
        """
        Sets name of configuration file associated with the plugin - note that setting this does not change/reload the
        configuration

        Args:
            filename (str): The name of configuration file associated with the plugin

        Returns:
            None
        """
        assert PanoptesValidators.valid_nonempty_string(filename), 'filename must be a non empty string'
        self._config_filename = filename

    @property
    def data(self):
        """
        The data associated with the plugin info object - this is passed to the plugin when it is being executed

        Returns:
            object: The data associated with the plugin info object
        """
        return self._data

    @data.setter
    def data(self, data):
        """
        Sets the data to be passed to the plugin when it is being executed

        Args:
            data (object): The data to be passed to the plugin when it is being executed

        Returns:
            None
        """
        assert PanoptesValidators.valid_hashable_object(data), 'plugin_data must be a valid hashable object'
        self._data = data

    @threaded_cached_property
    def signature(self):
        """
        The 'signature' of a plugin instance to uniquely identify it

        Returns:
            str: A hash of the plugin's config and data

        """
        return hashlib.md5(json.dumps(self.config, sort_keys=True) + str(hash(self.data))).hexdigest()

    @property
    def lock(self):
        """
        Attempts to get a lock for unique plugin instance

        Returns:
            KazooLock: The lock object

        Raises:
            Exception: Passes through any exceptions raised while trying to get a lock
        """
        if self._lock:
            return self._lock

        logger = self.panoptes_context.logger

        client_id = get_client_id(const.PLUGIN_CLIENT_ID_PREFIX)
        """
        We acquire a lock for a plugin under it's name and the hash of it's configuration and data. The motivation is
        that multiple instance of the plugins are allowed to execute in parallel - as long as they are acting
        on different resources or using different configurations
        """
        lock_path = '/'.join([const.PLUGIN_AGENT_LOCK_PATH,
                              self.normalized_category,
                              '/plugins/lock/',
                              self.normalized_name,
                              self.signature])

        logger.debug('Attempting to get lock for plugin "%s", with lock path "%s" and identifier "%s" in %d seconds' % (
            self.name, lock_path, client_id, const.PLUGIN_AGENT_LOCK_ACQUIRE_TIMEOUT))

        self._lock = PanoptesLock(context=self.panoptes_context,
                                  path=lock_path,
                                  timeout=const.PLUGIN_AGENT_LOCK_ACQUIRE_TIMEOUT,
                                  retries=1,
                                  identifier=client_id)

        return self._lock


class PanoptesBasePluginValidators(object):
    @classmethod
    def valid_plugin_class(cls, plugin_class):
        """
        valid_plugin_class(cls, plugin_class)

        Checks if the passed class is a subclass of PanoptesBasePlugin

        Args:
            plugin_class (cls): The class to check

        Returns:
            bool: True if the class is not null and is an subclass of PanoptesBasePlugin
        """
        return plugin_class and issubclass(plugin_class, PanoptesBasePlugin)


@six.add_metaclass(abc.ABCMeta)
class PanoptesBasePlugin(IPlugin):
    """
    The base class for all Panoptes plugins

    Every plugin in Panoptes needs to implement the 'run' method

    The Panoptes plugin system is based on `yapsy <http://yapsy.sourceforge.net/>`_

    """

    @abc.abstractmethod
    def run(self, context):
        """
        The entry point for every plugin in the Panoptes system

        Args:
            context (PanoptesPluginContext):  The context of the plugin contains the plugins configuration, the logger \
            it should use and a client for the system wide plugins KV store

        Returns:
            varies: The output of a plugin depends on the type of the plugin

        Raises:
            PanoptesPluginConfigurationError: If the plugin cannot proceed with the configuration provided to it
            through the context, it should raise a PanoptesPluginConfigurationError
        """
        pass



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This module implements the polling Plugin Agent which accepts plugin names as Celery Task parameters and executes
them. Results, if any, returned by each plugin are expected to be PanoptesMetricSets. If the returned Metrics Set is
not empty, each metric from the metric set is placed on a Kafka queue named 'metric'

This module is expected to be imported and executed though the Celery 'worker' command line tool
"""
import faulthandler
import sys

from celery import task
from celery.signals import worker_process_shutdown

from ..framework import const
from ..framework.context import PanoptesContext
from ..framework.exceptions import PanoptesBaseException
from ..framework.celery_manager import PanoptesCeleryConfig, PanoptesCeleryInstance
from ..framework.metrics import PanoptesMetricType, PanoptesMetric, PanoptesMetricsGroup, \
    PanoptesMetricsGroupSet
from ..framework.resources import PanoptesResourceStore, PanoptesResourcesKeyValueStore
from ..framework.plugins.panoptes_base_plugin import PanoptesPluginInfo
from ..framework.plugins.runner import PanoptesPluginWithEnrichmentRunner
from ..framework.utilities.helpers import get_calling_module_name
from ..framework.utilities.key_value_store import PanoptesKeyValueStore
from ..framework.utilities.secrets import PanoptesSecretsStore
from .polling_plugin import PanoptesPollingPlugin

panoptes_context = None
panoptes_polling_task_context = None
celery = None


class PanoptesMetricsKeyValueStore(PanoptesKeyValueStore):
    """Class to create key value store for metrics processing (particularly data to calculate gauges)

    Args:
        panoptes_context (PanoptesContext): The metrics panoptes context
    """
    redis_group = const.METRICS_REDIS_GROUP

    def __init__(self, panoptes_context):
        super(PanoptesMetricsKeyValueStore, self).__init__(panoptes_context, const.METRICS_KEY_VALUE_NAMESPACE)


class PanoptesPollingPluginAgentKeyValueStore(PanoptesKeyValueStore):
    """
    A custom Key/Value store for the Polling Plugin Agent which store data like a plugin's last execution time and \
    the last result production time
    """

    def __init__(self, panoptes_context):
        super(PanoptesPollingPluginAgentKeyValueStore, self).__init__(panoptes_context,
                                                                      const.POLLING_PLUGIN_AGENT_KEY_VALUE_NAMESPACE)


class PanoptesPollingPluginKeyValueStore(PanoptesKeyValueStore):
    """
    A custom Key/Value store for the Panoptes Polling plugins

    Plugins executed by the Polling Plugin Agent get access to this store and can get/set values they need for their\
    operation
    """

    def __init__(self, panoptes_context):
        super(PanoptesPollingPluginKeyValueStore, self).__init__(panoptes_context, const.PLUGINS_KEY_VALUE_NAMESPACE)


class PanoptesPollingAgentContext(PanoptesContext):
    """
    This class implements a PanoptesContext with no clients - only used for setting up the Celery app
    """

    def __init__(self):
        super(PanoptesPollingAgentContext, self).__init__(key_value_store_class_list=[],
                                                          create_message_producer=False,
                                                          create_zookeeper_client=False)


class PanoptesPollingTaskContext(PanoptesContext):
    """
    This class implements a PanoptesContext with the following:
        - A Key/Value Store for Polling Plugin Agent
        - A Key/Value Store for Panoptes Plugins
        - A Asynchronous Message Producer
        - A Zookeeper Client
    """

    def __init__(self):
        super(PanoptesPollingTaskContext, self).__init__(
                key_value_store_class_list=[PanoptesPollingPluginAgentKeyValueStore,
                                            PanoptesPollingPluginKeyValueStore,
                                            PanoptesSecretsStore,
                                            PanoptesResourcesKeyValueStore,
                                            PanoptesMetricsKeyValueStore],
                create_message_producer=True, async_message_producer=True, create_zookeeper_client=True)


class PanoptesPollingPluginAgentError(PanoptesBaseException):
    """
    The exception class for Polling Plugin Agent runtime errors
    """
    pass


@task
def polling_plugin_task(polling_plugin_name, resource_key):
    """
    The main method of the Polling Plugin Agent

    This method, called by Celery, loads and executes the specified plugin. The workflow is as follows:
        - Locate and load plugin through `yapsy's <http://yapsy.sourceforge.net/>`_ plugin manager
        - Evaluate if the plugin should be executed right now
        - If yes, then create a plugin context
        - Attempt to get a lock for the instance of the plugin (unique combination of the plugin name and configuration)
        - If the lock is acquired, then attempt to execute the plugin
        - If the plugin executes successfully, update the timestamp in the Polling Plugin Agent Key/Value store
        - If the plugin produced a non-zero result set, send the resource over the message bus
        - If sending over the message succeeds, update the timestamp in the Polling Plugin Agent Key/Value store

    Args:
        polling_plugin_name (str): The name of the plugin to be executed
        resource_key (dict): The dictionary that contains the resource key of the resource this polling plugin should \
        monitor

    Returns:
        None
    """
    global panoptes_polling_task_context

    if panoptes_polling_task_context is None:
        try:
            panoptes_polling_task_context = PanoptesPollingTaskContext()
        except Exception as e:
            sys.exit('Could not create a Panoptes Polling Task Context: %s' % (repr(e)))

    logger = panoptes_polling_task_context.logger

    try:
        resource = PanoptesResourceStore(panoptes_polling_task_context).get_resource(resource_key)
        plugin_runner = PanoptesPluginWithEnrichmentRunner(
                plugin_name=polling_plugin_name,
                plugin_type='polling',
                plugin_class=PanoptesPollingPlugin,
                plugin_info_class=PanoptesPluginInfo,
                plugin_data=resource,
                panoptes_context=panoptes_polling_task_context,
                plugin_agent_kv_store_class=PanoptesPollingPluginAgentKeyValueStore,
                plugin_kv_store_class=PanoptesPollingPluginKeyValueStore,
                plugin_secrets_store_class=PanoptesSecretsStore,
                plugin_logger_name=const.POLLING_PLUGIN_AGENT_PLUGINS_CHILD_LOGGER_NAME,
                plugin_result_class=PanoptesMetricsGroupSet,
                results_callback=_process_metrics_group_set
        )
        plugin_runner.execute_plugin()
    except Exception as e:
        logger.error('[%s] Error executing plugin: %s' % (polling_plugin_name, str(e)))


def _make_key(metrics_group):
    """
    Returns a key that is based on the resource id, metrics group type and dimension names and values

    Args:
        metrics_group (PanoptesMetricsGroup): The PanoptesMetricsGroup to generate the key for

    Returns:
        str: The KV_STORE_DELIMITER delimited key
    """
    dimensions = const.KV_STORE_DELIMITER.join(
            [const.KV_STORE_DELIMITER.join([dimension.name, dimension.value]) for dimension in
             metrics_group.dimensions])

    return const.KV_STORE_DELIMITER.join([metrics_group.resource.resource_id,
                                          metrics_group.group_type,
                                          dimensions])


def _split_and_strip(values, delimiter=','):
    return [value.strip() for value in values.split(delimiter)]


def _transformation_rate(context, metrics_group, inputs):
    kv_store = context.get_kv_store(PanoptesMetricsKeyValueStore)
    logger = context.logger

    output_metrics_group = metrics_group.copy()

    for metric in metrics_group.metrics:
        if metric.metric_name in inputs:
            key = const.KV_STORE_DELIMITER.join([_make_key(metrics_group), metric.metric_name])

            try:
                new_stored_value = const.KV_STORE_DELIMITER.join([str(metric.metric_value),
                                                                  str(metric.metric_timestamp)])
                stored_value = kv_store.getset(key, new_stored_value,
                                               const.METRICS_KV_STORE_TTL_MULTIPLE * metrics_group.interval)
            except Exception as e:
                context.logger.error(
                        'Error trying to fetch/store/convert for key "%s": %s, skipping conversion' % (key, repr(e)))
                continue

            if stored_value is None:
                logger.debug('Could not find existing value for key "%s", skipping conversion' % key)
                continue

            logger.debug('Calculating rate for %s' % key)

            value, timestamp = stored_value.split(const.KV_STORE_DELIMITER)
            time_difference = metric.metric_timestamp - float(timestamp)

            if time_difference < 0:
                logger.debug('Time difference is negative for key "%s": (%.2f), skipping conversion' %
                             (key, time_difference))
                continue
            elif time_difference == 0:
                logger.debug('Time difference is zero for key "%s", skipping conversion' % key)
                continue
            elif time_difference > (metrics_group.interval * const.METRICS_KV_STORE_TTL_MULTIPLE):
                logger.debug('Time difference is greater than TTL multiple for key "%s": (%.2f), skipping conversion' %
                             (key, time_difference))
                continue

            confidence = round(metrics_group.interval / time_difference, 2)
            if confidence < const.METRICS_CONFIDENCE_THRESHOLD:
                logger.warn('Confidence for key "%s" is %.2f, which is below the threshold of %.2f' %
                            (key, confidence, const.METRICS_CONFIDENCE_THRESHOLD))

            value = float(value)
            counter_difference = metric.metric_value - value

            if counter_difference >= 0:
                rate = int(counter_difference / time_difference)
                logger.debug('Rate for %s is %d' % (key, rate))

                try:
                    output_metrics_group.add_metric(PanoptesMetric(metric.metric_name, rate, PanoptesMetricType.GAUGE))
                except KeyError:
                    logger.warn('Metric %s already present as gauge in %s, skipping' % (metric.metric_name,
                                                                                        metrics_group.type))
            else:
                logger.debug('New counter value (%.2f) is less than current counter (%.2f) for key "%s": (%.2f), '
                             'skipping conversion' % (metric.metric_value, value, key, counter_difference))

    return output_metrics_group


def _process_transforms(context, transforms, metrics_group_set):
    """
    {
     "member": [{"rate": ["real_bytes_in",
                          "real_bytes_out",
                          "real_packets_in",
                          "real_packets_out",
                          "real_total_connections"]}]
    }

    Args:
        context (PanoptesContext): The PanoptesContext being used by the Plugin Agent
        transforms (dict): The transformations to apply
        metrics_group_set (PanoptesMetricsGroupSet): The metrics group set on which to apply the transformations

    Returns:
        PanoptesMetricsGroupSet: The processed/transformed metrics group set
    """
    callbacks = {'rate': _transformation_rate}

    logger = context.logger
    lookup = dict()
    output_metrics_group_set = PanoptesMetricsGroupSet()

    logger.debug('Going to process transforms: %s' % transforms)

    for key in transforms:
        transform_type, transform_metrics_group_type, transform_inputs = transforms[key].split(':')
        transform_inputs = _split_and_strip(transform_inputs)
        if transform_metrics_group_type not in lookup:
            lookup[transform_metrics_group_type] = list()

            lookup[transform_metrics_group_type].append((transform_type, transform_inputs))

    logger.debug('Transform lookups: %s' % lookup)

    for metrics_group in metrics_group_set:
        if metrics_group.group_type not in lookup:
            output_metrics_group_set.add(metrics_group)
            continue

        resource_serialization_key = metrics_group.resource.serialization_key

        for transform in lookup[metrics_group.group_type]:
            logger.debug(
                    'For resource %s, trying to process transform %s' % (resource_serialization_key, transform))
            transform_type, transform_inputs = transform

            if transform_type not in callbacks:
                logger.warn('For resource %s, no implementation for transform type "%s" found, skipping' % (
                    resource_serialization_key, transform))
                continue

            try:
                output_metrics_group = callbacks[transform_type](context, metrics_group, transform_inputs)
                if output_metrics_group is not None:
                    output_metrics_group_set.add(output_metrics_group)
            except Exception as e:
                logger.error('For resource %s, error while trying to transform metrics group "%s": %s, skipping' % (
                    resource_serialization_key, metrics_group.group_type, repr(e)))

    return output_metrics_group_set


def _send_metrics_group_set(context, metrics_group_set, topic_suffixes):
    """
    Emits the provided metrics group set to the message bus

    Args:
        context (PanoptesContext): The PanoptesContext being used by the Plugin Agent
        topic_suffixes (list): The list of suffixes to add to the topic to emit the metrics to
        metrics_group_set (PanoptesMetricsGroupSet): The metrics group set to emit

    Returns:
        None
    """
    producer = context.message_producer

    for metrics_group in metrics_group_set:
        key = const.METRICS_TOPIC_KEY_DELIMITER.join([metrics_group.resource.resource_class,
                                                      metrics_group.resource.resource_subclass,
                                                      metrics_group.resource.resource_type
                                                      ])
        partitioning_key = _make_key(metrics_group)

        for topic_suffix in set(topic_suffixes):
            topic = const.METRICS_TOPIC_NAME_DELIMITER.join([str(metrics_group.resource.resource_site), topic_suffix])
            context.logger.debug(
                    'Going to send metric group "%s" to topic "%s" with key "%s" and partitioning key "%s" ' % (
                        metrics_group.json, topic, key, partitioning_key))

            producer.send_messages(topic=topic, key=key, messages=metrics_group.json, partitioning_key=partitioning_key)

            context.logger.debug(
                    'Sent metric group "%s" to topic "%s" with key "%s" and partitioning key "%s" ' % (
                        metrics_group.json, topic, key, partitioning_key))


def _process_metrics_group_set(context, results, plugin):
    """
    Processes each metrics group in the result set - emits to the message bus, applies transforms and emits the \
    transformed results to the 'processed' topic on the message bus

    Args:
        context (PanoptesContext): The PanoptesContext being used by the Plugin Agent
        results (PanoptesMetricsGroupSet): The PanoptesMetricsGroupSet returned by the plugin
        plugin (PanoptesPollingPlugin): The Panoptes plugin object for the plugin that returned this dataset

    Returns:
        None
    """
    raw_metrics_topics_suffixes = [const.METRICS_RAW_TOPIC_SUFFIX]
    processed_metrics_topics_suffixes = [const.METRICS_PROCESSED_TOPIC_SUFFIX]

    if 'topics' in plugin.config:
        if 'raw' in plugin.config['topics']:
            raw_metrics_topics_suffixes.extend(_split_and_strip(plugin.config['topics']['raw']))
        if 'processed' in plugin.config['topics']:
            processed_metrics_topics_suffixes.extend(_split_and_strip(plugin.config['topics']['processed']))

    # Send to the 'raw' metrics topic
    _send_metrics_group_set(context, results, raw_metrics_topics_suffixes)

    # Applies transforms
    if 'transforms' in plugin.config:
        results = _process_transforms(context, plugin.config['transforms'], results)
        # Send to the 'processed' topic. Note - if no transforms have been applied,
        # then is essentially a copy of the 'raw' metrics
        _send_metrics_group_set(context, results, processed_metrics_topics_suffixes)


@worker_process_shutdown.connect()
def shutdown_signal_handler(sender, args=None, **kwargs):
    """
    This method handles the shutdown signal from Celery

    Args:
        sender (Celery): The Celery app which sent the shutdown signal
        args (dict): Unused
        **kwargs (dict): Unused

    Returns:
        None
    """
    if panoptes_polling_task_context is not None:
        print('Polling Plugin Agent shutting down - going to stop message producer')
        try:
            panoptes_polling_task_context.message_producer.stop()
        except Exception as e:
            print('Could not shutdown message producer: %s' % repr(e))


def start_polling_plugin_agent():
    """
    The entry point for the Polling Plugin Agent

    This method creates a Panoptes Context and the Celery Instance for the Polling Plugin Agent

    Returns:
        None
    """
    global panoptes_context, celery

    try:
        panoptes_context = PanoptesPollingAgentContext()
    except Exception as e:
        sys.exit('Could not create a Panoptes Context: %s' % (str(e)))

    logger = panoptes_context.logger
    logger.info('Attempting to start Celery application')

    celery_config = PanoptesCeleryConfig(const.POLLING_PLUGIN_AGENT_CELERY_APP_NAME)

    try:
        celery = PanoptesCeleryInstance(panoptes_context, celery_config).celery
    except Exception as exp:
        sys.exit('Could not instantiate Celery application: %s' % str(exp))
    else:
        logger.info('Started Celery application: %s' % celery)


if get_calling_module_name() == const.CELERY_LOADER_MODULE:
    faulthandler.enable()
    start_polling_plugin_agent()



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.
"""

class MockPanoptesLock(object):
    @staticmethod
    def locked():
        return True



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.
"""

from .. import const
from ..plugins.panoptes_base_plugin import PanoptesPluginInfo


def expires(plugin):
    """
    Returns the number of seconds after which to expire a pending plugin task

    Args:
        plugin (PanoptesPluginInfo): The plugin for which to return the expiry time

    Returns:
        int: The number of seconds after which to consider the task expired
    """
    assert isinstance(plugin, PanoptesPluginInfo), 'plugin must be an instance of PanoptesPluginInfo'

    return round(plugin.execute_frequency * const.PLUGIN_AGENT_PLUGIN_EXPIRES_MULTIPLE)


def time_limit(plugin):
    """
    Returns the number of seconds after which to stop a running plugin task

    Args:
        plugin (PanoptesPluginInfo): The plugin for which to return the time limit

    Returns:
        int: The number of seconds after which to stop a running plugin task
    """
    assert isinstance(plugin, PanoptesPluginInfo), 'plugin must be an instance of PanoptesPluginInfo'

    return round(plugin.execute_frequency * const.PLUGIN_AGENT_PLUGIN_TIME_LIMIT_MULTIPLE)



import os


def get_test_conf_file():
    my_dir = os.path.dirname(os.path.realpath(__file__))
    panoptes_test_conf_file = os.path.join(my_dir, 'config_files/test_panoptes_config.ini')

    return my_dir, panoptes_test_conf_file



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.
"""

import json
import os
import time

from yahoo_panoptes.framework.utilities.consumer import PanoptesConsumerTypes, make_topic_names_for_all_sites, \
    CONSUMER_TYPE_NAMES


# Mocking this since the underlying functions used in the original are only available on Linux systems
def mock_get_client_id(prefix):
    return '_'.join([prefix, 'localhost', '1234'])


class MockPanoptesConsumer(object):
    """
    A mock consumer for testing purposes
    """
    files = []

    def __init__(self, panoptes_context, consumer_type, topics, client_id, group, keys, poll_timeout, callback,
                 validate=False, session_timeout=10, max_poll_records=500, max_partition_fetch_bytes=1048576):
        self._panoptes_context = panoptes_context
        self._topics = topics
        self._client_id = client_id
        self._keys = keys
        self._group = group
        self._poll_timeout = poll_timeout * 1000
        self._callback = callback
        self._validate = validate

    def start_consumer(self):
        """
        Mock start consumer method uses static json object

        Note:
            Calls back based on _callback attribute passing args 'key:foo:foo' and the unpacked resources.json
        """
        for file_name in self.files:
            full_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), file_name)
            with open(full_path) as f:
                consumer_record_object = json.load(f)

                # If the timestamp is negative, then convert it to a relative timestamp by adding the current time
                # Skip for resource objects

                try:
                    if consumer_record_object['metrics_group_creation_timestamp'] < 0:
                        consumer_record_object['metrics_group_creation_timestamp'] += time.time()
                except:
                    pass

                self._callback('key:foo:foo', consumer_record_object)


class MockPanoptesResourcesConsumer(MockPanoptesConsumer):
    def __init__(self, panoptes_context, client_id, group, keys, poll_timeout, callback,
                 validate=False, session_timeout=10, max_poll_records=500, max_partition_fetch_bytes=1048576):

        topics = make_topic_names_for_all_sites(panoptes_context,
                                                CONSUMER_TYPE_NAMES[PanoptesConsumerTypes.RESOURCES])

        super(MockPanoptesResourcesConsumer, self).__init__(panoptes_context=panoptes_context,
                                                            consumer_type=PanoptesConsumerTypes.RESOURCES,
                                                            topics=topics,
                                                            client_id=client_id,
                                                            group=group,
                                                            keys=keys,
                                                            poll_timeout=poll_timeout,
                                                            callback=callback,
                                                            validate=validate,
                                                            session_timeout=session_timeout,
                                                            max_poll_records=max_poll_records,
                                                            max_partition_fetch_bytes=max_partition_fetch_bytes)



ifXTable = '.1.3.6.1.2.1.31.1.1.1'
ifName = ifXTable + '.1'
ifInMulticastPkts = ifXTable + '.2'
ifHCInOctets = ifXTable + '.6'
ifHCInUcastPkts = ifXTable + '.7'
ifHCInMulticastPkts = ifXTable + '.8'
ifHCInBroadcastPkts = ifXTable + '.9'
ifHCOutOctets = ifXTable + '.10'
ifHCOutUcastPkts = ifXTable + '.11'
ifHCOutMulticastPkts = ifXTable + '.12'
ifHCOutBroadcastPkts = ifXTable + '.13'
ifHighSpeed = ifXTable + '.15'
ifAlias = ifXTable + '.18'

ifx_table_oids = [ifHCInOctets,
                  ifHCInUcastPkts,
                  ifHCInMulticastPkts,
                  ifHCInBroadcastPkts,
                  ifHCOutOctets,
                  ifHCOutUcastPkts,
                  ifHCOutMulticastPkts,
                  ifHCOutBroadcastPkts,
                  ifHighSpeed,
                  ]



import inspect
import logging
from collections import Counter
from urllib2 import URLError

from requests.exceptions import ConnectTimeout, ConnectionError
from urllib3.exceptions import ConnectTimeoutError

from ....framework.metrics import PanoptesMetricsGroup, PanoptesMetric, PanoptesMetricType
from ....framework.metrics import PanoptesMetricsNullException
from ....framework.resources import PanoptesResource
from ....framework.utilities.ping import *
from ....framework.utilities.snmp.exceptions import *


class DEVICE_METRICS_STATES(object):
    """
    DEVICE_METRICS_STATES encapsulates all the states a device may have during the process of collecting or attempting
        to collect metrics from said device.
    """
    SUCCESS, \
        AUTHENTICATION_FAILURE, \
        NETWORK_FAILURE, \
        TIMEOUT, \
        PARTIAL_METRIC_FAILURE, \
        INTERNAL_FAILURE, \
        MISSING_METRICS, \
        PING_FAILURE = range(8)


exceptions_dict = {
    ConnectTimeout: DEVICE_METRICS_STATES.TIMEOUT,
    ConnectTimeoutError: DEVICE_METRICS_STATES.TIMEOUT,
    ConnectionError: DEVICE_METRICS_STATES.NETWORK_FAILURE,
    PanoptesSNMPTimeoutException: DEVICE_METRICS_STATES.TIMEOUT,
    PanoptesSNMPConnectionException: DEVICE_METRICS_STATES.NETWORK_FAILURE,
    URLError: DEVICE_METRICS_STATES.NETWORK_FAILURE,
    PanoptesPingException: DEVICE_METRICS_STATES.PING_FAILURE,
    PanoptesPingTimeoutException: DEVICE_METRICS_STATES.PING_FAILURE,
    PanoptesMetricsNullException: DEVICE_METRICS_STATES.MISSING_METRICS
}

EXCEPTIONS_KEYS = exceptions_dict.keys()

_PING_STATES = [DEVICE_METRICS_STATES.TIMEOUT, DEVICE_METRICS_STATES.NETWORK_FAILURE]


class PanoptesPollingStatus(object):
    """
    PanoptesPollingStatus is the primary class for tracking metrics collection for devices in the Panoptes environment.

    Args:
        resource(PanoptesResource): Which PanoptesResource this PanoptesPollingStatus instance will refer to
        execute_frequency(int): How frequently to execute the polling on the resource
        logger(logging.Logger): log for output
        ping(bool): Ping the device after any metric collection failure
        metrics_group_type_name(str): The metrics group type name, defaults to 'status'
        metric_name(str): The metric name, defaults to 'status'
    """

    def __init__(self, resource, execute_frequency, logger, ping=True,
                 metrics_group_type_name='status', metric_name='status'):
        assert isinstance(resource, PanoptesResource), 'resource must be an instance of PanoptesResource'
        assert PanoptesValidators.valid_nonzero_integer(execute_frequency), 'execute_frequency must be integer > 0'
        assert PanoptesValidators.valid_logger(logger), 'logger must be an instance of logging.Logger'

        self._device_name = resource.resource_endpoint
        self._device_type = ':'.join([resource.resource_class, resource.resource_subclass, resource.resource_type])
        self._device_status_metrics_group = PanoptesMetricsGroup(resource, metrics_group_type_name, execute_frequency)
        self._metric_name = metric_name
        self._logger = logger
        self._metric_statuses = dict()
        self._device_status = DEVICE_METRICS_STATES.SUCCESS
        self._ping = ping

    @property
    def device_status_metrics_group(self):
        """
        Creates device_status_metrics_group

        Returns:
            PanoptesMetricsGroup: The PanoptesMetricsGroup for the status of this device
        """
        if self._ping and self.device_status in _PING_STATES:
            try:
                panoptes_ping = PanoptesPing(hostname=self.device_name)
                if panoptes_ping.packet_loss_pct == 100.0:
                    self._device_status = DEVICE_METRICS_STATES.PING_FAILURE
            except:
                self._device_status = DEVICE_METRICS_STATES.PING_FAILURE

        self._device_status_metrics_group.add_metric(PanoptesMetric(self._metric_name,
                                                                    self._device_status,
                                                                    PanoptesMetricType.GAUGE))
        return self._device_status_metrics_group

    @property
    def device_status(self):
        """
        Returns:
            int: The status of the "device" metric
        """
        return self._device_status

    @property
    def device_name(self):
        """
        Returns:
             str: The name of this device
        """
        return self._device_name

    @property
    def device_type(self):
        """
        Returns:
             str: The type of this device
        """
        return self._device_type

    @property
    def logger(self):
        """
        Returns:
             logging.Logger: The logger for this device
        """
        return self._logger

    def handle_success(self, k):
        """
        Update the _device_metrics_status dictionary when a given metric has been successfully obtained.

        Args:
            k(str): name of the metric to apply the success to
        """
        self.logger.debug(
                'Successfully polled %s "%s" for %s' % (str(self._device_type), str(self._device_name), str(k)))
        if k in self._metric_statuses:
            if self._metric_statuses[k] != DEVICE_METRICS_STATES.SUCCESS:
                self._metric_statuses[k] = DEVICE_METRICS_STATES.PARTIAL_METRIC_FAILURE
        else:
            self._metric_statuses[k] = DEVICE_METRICS_STATES.SUCCESS
        self._set_device_status()

    def handle_exception(self, k, e):
        """
        Mutate the state of the _device_metrics_status dictionary.

        Args:
            k(str): name of the metric to apply the exception to
            e(Exception): exception which has occurred for the given k
        """
        self.logger.warn('Error while trying to poll "%s" (%s) for "%s": %s' %
                         (str(self._device_name), str(self._device_type), k, repr(e)))

        if k in self._metric_statuses:
            if self._metric_statuses[k] in \
                    [DEVICE_METRICS_STATES.SUCCESS, DEVICE_METRICS_STATES.PARTIAL_METRIC_FAILURE]:
                self._metric_statuses[k] = DEVICE_METRICS_STATES.PARTIAL_METRIC_FAILURE
                self._set_device_status()
                return

        if type(e) in EXCEPTIONS_KEYS:
            self._metric_statuses[k] = exceptions_dict[type(e)]
        else:
            found_exception = False
            for exception in inspect.getmro(type(e)):
                if exception in EXCEPTIONS_KEYS:
                    self._metric_statuses[k] = exceptions_dict[exception]
                    found_exception = True
                    break
            if not found_exception:
                self._metric_statuses[k] = DEVICE_METRICS_STATES.INTERNAL_FAILURE
        self._set_device_status()

    def _set_device_status(self):
        """
        Call when done with operations on polling status object to set overall
        device status based upon component metric statuses.
        """
        if len(self._metric_statuses) > 0:
            if all(status == DEVICE_METRICS_STATES.SUCCESS for status in self._metric_statuses.values()):
                self._device_status = DEVICE_METRICS_STATES.SUCCESS
            elif DEVICE_METRICS_STATES.SUCCESS in self._metric_statuses.values():
                self._device_status = DEVICE_METRICS_STATES.PARTIAL_METRIC_FAILURE
            else:
                count = Counter(self._metric_statuses.values())
                if len(count.most_common()) > 0:
                    # get the most common (1) status and it's count as a tuple, and grab just its name
                    self._device_status = count.most_common(1)[0][0]
                else:
                    self._device_status = DEVICE_METRICS_STATES.INTERNAL_FAILURE
